# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import glob
import os
import pathlib
import shutil
import subprocess
import tkinter


class ProfilingContainer(object):
    def __init__(self, wizard):
        self.wizard = wizard
        self.start()

    def start(self):
        docker_context_path = os.path.join(
            pathlib.Path(__file__).parent.resolve(), "..", "assets", "profiling_container"
        )
        self.__execute_command("docker kill discopop_container")
        self.__execute_command("docker rm discopop_container")
        exit_code = self.__execute_command("docker build -t discopop_container " + docker_context_path)
        assert exit_code == 0
        exit_code = self.__execute_command("docker run --name discopop_container -d -t discopop_container")
        assert exit_code == 0

    def stop(self):
        print("Stopping DiscoPoP profiling container...")
        self.__execute_command("docker kill discopop_container")
        print("Done.")

    def remove_project_folder(self):
        self.__execute_command("docker exec -it discopop_container rm -rvf /project")

    def remove_previous_results(self, target_folder):
        files = glob.glob(target_folder + "/*")
        for f in files:
            if os.path.exists(f):
                if os.path.isfile(f):
                    os.remove(f)
                else:
                    shutil.rmtree(f)

    def copy_project_folder_to_container(self, project_path: str):
        self.remove_project_folder()
        self.__execute_command("docker cp " + project_path + " discopop_container:/project")

    def copy_results_from_container(self, target_path: str, execution_view):
        result_files = [
            "FileMapping.txt",
            "Data.xml",
            "loop_counter_output.txt",
            "reduction.txt",
            execution_view.execution_configuration.value_dict["executable_name"] + "_dp_dep.txt",
            execution_view.execution_configuration.value_dict["executable_name"] + "_dp.ll",
            "patterns.txt",
            "patterns.json",
        ]
        for file in result_files:
            exit_code = self.__execute_command(
                "docker cp discopop_container:/project/.discopop/" + file + " " + target_path
            )
            assert exit_code == 0

    def analyze_project(self, execution_view):
        # copy project folder to container. Note: mounting would be nicer but requires restarting the container.
        # might be a nicer solution in the long run, especially for larger projects
        self.copy_project_folder_to_container(execution_view.execution_configuration.value_dict["project_path"])

        # settings
        command = "/discopop/build/scripts/runDiscoPoP "
        command += "--llvm-clang clang-11 "
        command += "--llvm-clang++ clang++-11 "
        command += "--llvm-ar llvm-ar-11 "
        command += "--llvm-link llvm-link-11 "
        command += "--llvm-dis llvm-dis-11 "
        command += "--llvm-opt opt-11 "
        command += "--llvm-llc llc-11 "
        command += "--gllvm /software/go/bin "
        # execution configuration
        command += "--project /project "
        command += '--linker-flags "' + execution_view.execution_configuration.value_dict["linker_flags"] + '" '
        command += '--executable-name "' + execution_view.execution_configuration.value_dict["executable_name"] + '" '
        command += (
            '--executable-arguments "'
            + execution_view.execution_configuration.value_dict["executable_arguments"]
            + '" '
        )
        command += '--make-flags "' + execution_view.execution_configuration.value_dict["make_flags"] + '" '
        command += '--make-target "' + execution_view.execution_configuration.value_dict["make_target"] + '" '
        command += '--explorer-flags "' + execution_view.execution_configuration.value_dict["explorer_flags"] + '" '

        self.__execute_command("docker exec -it discopop_container " + command)

        # copy results from container into working copy path
        if not os.path.exists(execution_view.execution_configuration.value_dict["working_copy_path"]):
            os.mkdir(execution_view.execution_configuration.value_dict["working_copy_path"])

        # remove previous results
        self.remove_previous_results(execution_view.execution_configuration.value_dict["working_copy_path"])

        # copy results from container
        self.copy_results_from_container(
            execution_view.execution_configuration.value_dict["working_copy_path"], execution_view
        )

        # correct paths in generated FileMapping.txt
        self.__correct_file_mapping_paths(execution_view)

    def __correct_file_mapping_paths(self, execution_view):
        file_mapping_path = os.path.join(
            execution_view.execution_configuration.value_dict["working_copy_path"],
            "FileMapping.txt",
        )
        with open(file_mapping_path, "r") as file:
            contents = file.read()
        contents = contents.replace(
            "/project/.discopop", execution_view.execution_configuration.value_dict["project_path"]
        )
        with open(file_mapping_path, "w") as file:
            file.write(contents)

    def __execute_command(self, command: str) -> int:
        with subprocess.Popen(command, stdout=subprocess.PIPE, bufsize=1, universal_newlines=True, shell=True) as p:
            if p.stdout is None:
                print("executing command was not successfull")
            else:
                for line in p.stdout:
                    line = line.replace("\n", "")
                    print(line)
                    try:
                        self.wizard.console.print(line)
                    except tkinter.TclError:
                        # happens when container is still shutting down but interface already closed.
                        pass
        if p.returncode != 0:
            print("An error occurred during the execution!")  # Error message
            self.wizard.console.print("An error occurred during the execution!")
            for line in str(subprocess.CalledProcessError(p.returncode, p.args)).split("\n"):
                line = line.replace("\n", "")
                print(line)
                self.wizard.console.print(line)
        return p.returncode



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import copy
from typing import Set, cast, Tuple, List, Dict

import networkx as nx  # type: ignore
from sympy import Expr, Integer, Symbol, log, Float, init_printing  # type: ignore

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.Microbench.utils import (
    convert_discopop_to_microbench_workload,
)
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.edges.MutuallyExclusiveEdge import MutuallyExclusiveEdge
from discopop_library.discopop_optimizer.classes.edges.OptionEdge import OptionEdge
from discopop_library.discopop_optimizer.classes.nodes.Loop import Loop
from discopop_library.discopop_optimizer.classes.nodes.Workload import Workload
from discopop_library.discopop_optimizer.classes.system.devices.CPU import CPU
from discopop_library.discopop_optimizer.classes.system.devices.GPU import GPU
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at

suggestion_device_types = [CPU, GPU]


def import_suggestion(
    graph: nx.DiGraph, suggestion, get_next_free_node_id_function, environment: Experiment
) -> nx.DiGraph:
    # find a node which belongs to the suggestion
    buffer = [n for n in graph.nodes]
    # get all devices which can be used to execute the suggestion
    suggestion_device_ids = []
    for device_type in suggestion_device_types:
        suggestion_device_ids += environment.get_system().get_device_ids_by_type(device_type)

    for node in buffer:
        introduced_options: Set[int] = set()
        if suggestion.node_id == data_at(graph, node).cu_id:
            # save node in introduced_options to mark as mutually exclusive
            introduced_options.add(node)
            for device_id in suggestion_device_ids:
                # reserve a node id for the new parallelization option
                new_node_id = get_next_free_node_id_function()
                # copy data from existing node
                node_data_copy = copy.deepcopy(data_at(graph, node))
                node_data_copy.node_id = new_node_id
                # register new node_id for pattern
                if suggestion.pattern_id not in environment.suggestion_to_node_ids_dict:
                    environment.suggestion_to_node_ids_dict[suggestion.pattern_id] = []
                environment.suggestion_to_node_ids_dict[suggestion.pattern_id].append(new_node_id)
                environment.node_id_to_suggestion_dict[new_node_id] = suggestion.pattern_id

                # set the device id for the suggestion
                node_data_copy.device_id = device_id
                # remove cu_id to prevent using parallelization options as basis for new versions
                node_data_copy.cu_id = None
                # mark node for parallel execution
                node_data_copy.execute_in_parallel = True
                # copy loop iteration variable
                cast(Loop, node_data_copy).iterations_symbol = cast(Loop, node_data_copy).iterations_symbol
                # add suggestion to node data
                node_data_copy.suggestion = suggestion
                node_data_copy.suggestion_type = "reduction"
                # add the cost multiplier to represent the effects of the suggestion
                (
                    cast(Workload, node_data_copy).cost_multiplier,
                    introduced_symbols,
                ) = get_cost_multiplier(new_node_id, environment, device_id)
                # add the overhead term to represent the overhead incurred by the suggestion
                cast(Workload, node_data_copy).overhead, tmp_introduced_symbols = get_overhead_term(
                    cast(Loop, node_data_copy), environment, device_id
                )
                introduced_symbols += tmp_introduced_symbols

                node_data_copy.introduced_symbols += introduced_symbols

                # create a new node for the option
                graph.add_node(new_node_id, data=node_data_copy)
                # mark the newly created option
                # graph.add_edge(node, new_node_id, data=OptionEdge())

                # save the id of the introduced parallelization option to connect them afterwards
                introduced_options.add(new_node_id)

                # connect the newly created node to the parent and successor of node
                for edge in graph.in_edges(node):
                    edge_data = copy.deepcopy(graph.edges[edge]["data"])
                    graph.add_edge(edge[0], new_node_id, data=edge_data)
                for edge in graph.out_edges(node):
                    edge_data = copy.deepcopy(graph.edges[edge]["data"])
                    graph.add_edge(new_node_id, edge[1], data=edge_data)
                    # if the successor has no device id already,
                    # set it to 0 to simulate "leaving" the device after the suggestion
                    # todo: this should not happen here, but be considered when calculating the updates in order to
                    #  prevent suggestions from influencing each other by "mapping" workloads to certain devices.
                    # todo re-enable?
                    # if data_at(graph, edge[1]).device_id is None:
                    #     data_at(graph, edge[1]).device_id = 0

        # connect introduced parallelization options to support path restraining
        for node_id_1 in introduced_options:
            for node_id_2 in introduced_options:
                if node_id_1 == node_id_2:
                    continue
                graph.add_edge(node_id_1, node_id_2, data=MutuallyExclusiveEdge())
    return graph


def get_cost_multiplier(node_id: int, environment: Experiment, device_id: int) -> Tuple[CostModel, List[Symbol]]:
    """Creates and returns the multiplier to represent the effects of the given suggestion on the cost model.
    A CostModel object is used to store the information on the path selection.
    Returns the multiplier and the list of introduces symbols
    Multiplier for Reduction:
        1 / Compute_capa"""
    # get device specifications
    thread_count = environment.get_system().get_device(device_id).get_thread_count()

    multiplier = Integer(1) / thread_count
    cm = CostModel(multiplier, Integer(1))

    # return cm, [thread_count]
    return cm, []


def get_overhead_term(node_data: Loop, environment: Experiment, device_id: int) -> Tuple[CostModel, List[Symbol]]:
    """Creates and returns the Expression which represents the Overhead incurred by the given suggestion.
    For testing purposes, the following function is used to represent the overhead incurred by a do-all loop.
    The function has been created using Extra-P.
    Unit of the overhead term are micro seconds."""

    # retrieve Reduction overhead model
    overhead_model = environment.get_system().get_device_reduction_overhead_model(
        environment.get_system().get_device(device_id), environment.arguments
    )
    # substitute workload, iterations and threads
    thread_count = environment.get_system().get_device(device_id).get_thread_count()
    iterations = node_data.iterations_symbol
    # since node_data is of type Loop, parallelizable_workload has to exist
    per_iteration_workload = cast(Expr, node_data.parallelizable_workload)
    # convert DiscoPoP workload to Microbench workload
    converted_per_iteration_workload = convert_discopop_to_microbench_workload(per_iteration_workload, iterations)

    substitutions: Dict[Symbol, Expr] = {}

    for symbol in cast(List[Symbol], overhead_model.free_symbols):
        if symbol.name == "workload":
            substitutions[symbol] = converted_per_iteration_workload
        elif symbol.name == "iterations":
            substitutions[symbol] = iterations
        elif symbol.name == "threads":
            substitutions[symbol] = thread_count
        else:
            raise ValueError("Unknown symbol: ", symbol)

    substituted_overhead_model = overhead_model.xreplace(substitutions)

    cm = CostModel(Integer(0), substituted_overhead_model)

    # todo: convert result (in s) to workload

    # add weight to overhead
    return cm, []



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
from dataclasses import dataclass
from typing import Dict, List, Tuple

from .Microbench import (
    MicrobenchType,
    MicrobenchCoordinate,
    MicrobenchDimension,
)

# threads/workload/iterations
ParsedCoordinate = Tuple[int, int, int]


@dataclass
class ParsedMicrobenchResults:
    measurements: Dict[
        MicrobenchType,
        Dict[MicrobenchDimension, Dict[MicrobenchCoordinate, List[float]]],
    ]


def parseMicrobenchResults(jsonPath: str) -> ParsedMicrobenchResults:
    # Parse JSON into an object with attributes corresponding to dict keys.
    with open(jsonPath) as file:
        jsonData = json.loads(file.read())

    measurements: Dict[
        MicrobenchType,
        Dict[MicrobenchDimension, Dict[MicrobenchCoordinate, List[float]]],
    ] = dict()
    for typeString, dimMap in jsonData["measurements"].items():
        mbType = MicrobenchType(typeString)
        measurements[mbType] = dict()
        for dimString, coordList in dimMap.items():
            mbDim = MicrobenchDimension(dimString)
            measurements[mbType][mbDim] = dict()
            for obj in coordList:
                coord = MicrobenchCoordinate(*obj["point"])
                values = obj["values"]
                measurements[mbType][mbDim][coord] = values
    return ParsedMicrobenchResults(measurements)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import Set

from discopop_explorer.PEGraphX import NodeID, MemoryRegion
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Aliases import (
    VarName,
)


class Dependency(object):
    source: NodeID
    sink: NodeID
    var_names: Set[VarName]
    memory_regions: Set[MemoryRegion]

    def __init__(
        self,
        source: NodeID,
        sink: NodeID,
        var_names: Set[VarName],
        memory_regions: Set[MemoryRegion],
    ):
        self.source = source
        self.sink = sink
        self.var_names = var_names
        self.memory_regions = memory_regions

    def __str__(self):
        return (
            "Dependency("
            + self.source
            + " -> "
            + self.sink
            + " "
            + str(self.var_names)
            + " / "
            + str(self.memory_regions)
            + ")"
        )



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Dict, Tuple, List, Optional, cast
import warnings

from sympy import Expr, Symbol, Integer

from discopop_library.discopop_optimizer.classes.system.devices.Device import Device


class Network(object):
    __host_device: Optional[Device]
    __transfer_speeds: Dict[Tuple[Device, Device], Expr]  # (MB/s)
    __transfer_initialization_costs: Dict[Tuple[Device, Device], Expr]

    def __init__(self):
        self.__host_device = None
        self.__transfer_speeds = dict()
        self.__transfer_initialization_costs = dict()

    def add_connection(self, source: Device, target: Device, transfer_speed: Expr, initialization_delay: Expr):
        if source == target:
            transfer_speed = Integer(10000 * 1000)  # 10000 GB/s = 10000 * 1000 MB/s
        self.__transfer_speeds[(source, target)] = transfer_speed
        self.__transfer_initialization_costs[(source, target)] = initialization_delay

    def get_transfer_speed(self, source: Device, target: Device):
        if (source, target) not in self.__transfer_speeds:
            if self.__host_device is None:
                raise ValueError("Host device of network unspecified!")
            S_to_H_speed = self.__transfer_speeds[(source, self.__host_device)]
            H_to_T_speed = self.__transfer_speeds[(self.__host_device, target)]
            return min(S_to_H_speed, H_to_T_speed)
        return self.__transfer_speeds[(source, target)]

    def get_transfer_initialization_costs(self, source: Device, target: Device):
        if (source, target) not in self.__transfer_speeds:
            if self.__host_device is None:
                raise ValueError("Host device of network unspecified!")
            S_to_H_init_costs = self.__transfer_initialization_costs[(source, self.__host_device)]
            H_to_T_init_costs = self.__transfer_initialization_costs[(self.__host_device, target)]
            return S_to_H_init_costs + H_to_T_init_costs
        return self.__transfer_initialization_costs[(source, target)]

    def get_free_symbols(self) -> List[Tuple[Symbol, Optional[Expr]]]:
        result_list: List[Tuple[Symbol, Optional[Expr]]] = []
        for expr in self.__transfer_speeds.values():
            result_list += [(cast(Symbol, s), None) for s in expr.free_symbols]
        for expr in self.__transfer_initialization_costs.values():
            result_list += [(cast(Symbol, s), None) for s in expr.free_symbols]
        return result_list

    def set_host_device(self, host_device: Device):
        self.__host_device = host_device



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from __future__ import annotations

import copy
import itertools
from enum import IntEnum, Enum
from typing import Dict, List, Sequence, Tuple, Set, Optional, Type, TypeVar, cast, Union, overload, Any

import jsonpickle  # type:ignore
import matplotlib.pyplot as plt  # type:ignore
import networkx as nx  # type:ignore
from alive_progress import alive_bar  # type: ignore
from lxml.objectify import ObjectifiedElement  # type:ignore

from .parser import LoopData, readlineToCUIdMap, writelineToCUIdMap, DependenceItem
from .variable import Variable

global_pet = None

# unused
# node_props = [
#    ("BasicBlockID", "string", "''"),
#    ("pipeline", "float", "0"),
#    ("doAll", "bool", "False"),
#    ("geomDecomp", "bool", "False"),
#    ("reduction", "bool", "False"),
#    ("mwType", "int", "2"),
#    ("localVars", "object", "[]"),
#    ("globalVars", "object", "[]"),
#    ("args", "object", "[]"),
#    ("recursiveFunctionCalls", "object", "[]"),
# ]

# edge_props = [
#    ("type", "string"),
#    ("source", "string"),
#    ("sink", "string"),
#    ("var", "string"),
#    ("dtype", "string"),
# ]


def parse_id(node_id: str) -> Tuple[int, int]:
    split = node_id.split(":")
    return int(split[0]), int(split[1])


class EdgeType(Enum):
    CHILD = 0
    SUCCESSOR = 1
    DATA = 2
    CALLSNODE = 3
    PRODUCE_CONSUME = 4


class DepType(Enum):
    RAW = 0
    WAR = 1
    WAW = 2
    INIT = 3


class NodeType(IntEnum):
    CU = 0
    FUNC = 1
    LOOP = 2
    DUMMY = 3


class MWType(Enum):
    NONE = 0
    ROOT = 1
    FORK = 2
    WORKER = 3
    BARRIER = 4
    BARRIER_WORKER = 5


class NodeID(str):
    # simpler but still strong typing alternative:
    # NodeID = NewType("NodeID", str) or remove __init__
    def __init__(self, id_string: str):
        # check format of newly created NodeID's
        if ":" not in id_string:
            raise ValueError("Mal-formatted NodeID: ", id_string)
        split_id: List[str] = id_string.split(":")
        if len(split_id) != 2:
            raise ValueError("Mal-formatted NodeID: ", id_string)
        try:
            int(split_id[0])
            int(split_id[1])
        except ValueError:
            raise ValueError("Mal-formatted NodeID: ", id_string)


class LineID(str):
    # simpler but still strong typing alternative:
    # LineID = NewType("LineID", str) or remove __init__
    def __init__(self, id_string: str):
        # check format of newly created LineID's
        if ":" not in id_string:
            raise ValueError("Mal-formatted LineID: ", id_string)
        split_id: List[str] = id_string.split(":")
        if len(split_id) != 2:
            raise ValueError("Mal-formatted LineID: ", id_string)
        try:
            int(split_id[0])
            int(split_id[1])
        except ValueError:
            raise ValueError("Mal-formatted LineID: ", id_string)


class MemoryRegion(str):
    # simpler but still strong typing alternative:
    def __init__(self, id_string: str):
        super().__init__()
        pass


#        # check format of newly created MemoryRegion
#        try:
#            int(id_string)
#        except ValueError:
#            raise ValueError("Mal-formatted MemoryRegion identifier: ", id_string)


class Dependency:
    etype: EdgeType
    dtype: Optional[DepType] = None
    var_name: Optional[str] = None
    memory_region: Optional[MemoryRegion] = None
    source_line: Optional[LineID] = None
    sink_line: Optional[LineID] = None
    intra_iteration: bool = False
    intra_iteration_level: int = -1

    def __init__(self, type: EdgeType):
        self.etype = type

    def __str__(self):
        return self.var_name if self.var_name is not None else str(self.etype)


# TODO make the Node class abstract
class Node:
    # properties of every Node
    id: NodeID
    file_id: int
    node_id: int
    start_line: int
    end_line: int
    type: NodeType
    name: str
    parent_function_id: Optional[
        NodeID
    ] = None  # metadata to speedup some calculations (TODO FunctionNodes have themselves as parent)
    workload: Optional[int] = None

    # properties of CU Nodes
    node_calls: List[Dict[str, str]] = []
    recursive_function_calls: List[str] = []

    # properties related to pattern analysis
    reduction: bool = False
    do_all: bool = False
    geometric_decomposition: bool = False
    pipeline: float = -1
    tp_contains_task: bool = False
    tp_contains_taskwait: bool = False
    tp_omittable: bool = False
    mw_type = MWType.FORK

    def __init__(self, node_id: NodeID):
        self.id = node_id
        self.file_id, self.node_id = parse_id(node_id)

    @classmethod
    def from_kwargs(cls, node_id: NodeID, **kwargs):
        node = cls(node_id)
        for key, value in kwargs.items():
            setattr(node, key, value)
        return node

    def start_position(self) -> LineID:
        """Start position file_id:line
        e.g. 23:45

        :return:
        """

        return LineID(f"{self.file_id}:{self.start_line}")

    def end_position(self) -> LineID:
        """End position file_id:line
        e.g. 23:45

        :return:
        """
        return LineID(f"{self.file_id}:{self.end_line}")

    def contains_line(self, other_line) -> bool:
        if other_line == "GlobalVar":
            return False
        other_file_id = int(other_line.split(":")[0])
        other_line_num = int(other_line.split(":")[1])
        if other_file_id != self.file_id:
            return False
        if other_line_num >= self.start_line and other_line_num <= self.end_line:
            return True
        return False

    def __str__(self):
        return self.id

    def __eq__(self, other):
        return isinstance(other, Node) and other.file_id == self.file_id and other.node_id == self.node_id

    def __hash__(self):
        return hash(self.id)

    def get_parent_id(self, pet: PEGraphX) -> Optional[NodeID]:
        parents = [s for s, t, d in pet.in_edges(self.id, EdgeType.CHILD)]
        if len(parents) == 0:
            return None
        elif len(parents) == 1:
            return parents[0]
        else:
            # it is possible that a node has a function-type and e.g. loop type parent
            # in this case, return the non-function type parent, since it will be a child of the function itself.
            if len(parents) > 2:
                raise ValueError("Node: ", self.id, "has too many parents!")
            else:
                for parent in parents:
                    if type(pet.node_at(parent)) != FunctionNode:
                        return parent
        return None


# Data.xml: type="0"
class CUNode(Node):
    instructions_count: int = -1
    basic_block_id = ""
    return_instructions_count: int = -1
    local_vars: List[Variable] = []
    global_vars: List[Variable] = []
    performs_file_io: bool = False

    def __init__(self, node_id: NodeID):
        super().__init__(node_id)
        self.type = NodeType.CU


# Data.xml: type="2"
class LoopNode(Node):
    loop_iterations: int = -1
    contains_array_reduction: bool = False
    loop_data: Optional[LoopData]
    loop_indices: List[str]

    def __init__(self, node_id: NodeID):
        super().__init__(node_id)
        self.type = NodeType.LOOP
        self.loop_indices = []

    def get_nesting_level(self, pet: PEGraphX, return_invert_result: bool = True) -> int:
        """Returns the loop nesting level for the given loop node.
        Currently, due to the profiling output, only 3 nesting levels of loops can be mapped correctly.
        Innermost level is 0.
        Outermost (maximum) level is 2.
        The order of the results originates from the LoopStack used during the profiling.

        Example:
        main(){
            for(){ // level 1
                for(){}     // level 0
            }
        }
        """
        parents = [s for s, t, d in pet.in_edges(self.id, EdgeType.CHILD)]

        # count levels upwards
        parent_nesting_levels: List[int] = []
        for parent_id in parents:
            parent_node = pet.node_at(parent_id)
            if type(parent_node) == FunctionNode:
                # loop is a direct child of a function node --> Nesting level 0
                parent_nesting_levels.append(0)
                break
            elif type(parent_node) == LoopNode:
                parent_nesting_levels.append(
                    min(
                        2,
                        parent_node.get_nesting_level(pet, return_invert_result=False),
                    )
                )

        if return_invert_result:
            # invert the leveling and cutoff at 0
            inverted_levels = [max(0, 2 - level) for level in parent_nesting_levels]
            return min(inverted_levels)
        else:
            return max(parent_nesting_levels)

    def get_entry_node(self, pet: PEGraphX) -> Optional[Node]:
        """returns the first CU Node contained in the loop (i.e. one without predecessor inside the loop)"""
        for node in pet.direct_children(self):
            predecessors_outside_loop_body = [
                s
                for s, t, d in pet.in_edges(node.id, EdgeType.SUCCESSOR)
                if pet.node_at(s) not in pet.direct_children(self)
            ]
            if len(predecessors_outside_loop_body) > 0:
                return node
        return None


# Data.xml: type="3"
class DummyNode(Node):
    args: List[Variable] = []

    def __init__(self, node_id: NodeID):
        super().__init__(node_id)
        self.type = NodeType.DUMMY


# Data.xml: type="1"
class FunctionNode(Node):
    args: List[Variable] = []
    children_cu_ids: Optional[List[NodeID]] = None  # metadata to speedup some calculations
    reachability_pairs: Dict[NodeID, Set[NodeID]]
    immediate_post_dominators: Dict[NodeID, NodeID]
    immediate_post_dominators_present: bool
    memory_accesses: Dict[int, Dict[MemoryRegion, Set[Optional[int]]]]
    memory_accesses_present: bool

    def __init__(self, node_id: NodeID):
        super().__init__(node_id)
        self.type = NodeType.FUNC
        self.reachability_pairs = dict()
        self.immediate_post_dominators_present = False
        self.memory_accesses_present = False

    def get_entry_cu_id(self, pet: PEGraphX) -> NodeID:
        for child_cu_id in [t for s, t, d in pet.out_edges(self.id, EdgeType.CHILD)]:
            if len(pet.in_edges(child_cu_id, EdgeType.SUCCESSOR)) == 0:
                return child_cu_id
        raise ValueError("Mal-formatted function: ", self.id, " - No entry CU found!")

    def get_exit_cu_ids(self, pet: PEGraphX) -> Set[NodeID]:
        exit_cu_ids: Set[NodeID] = set()
        if self.children_cu_ids is not None:
            for child_cu_id in self.children_cu_ids:
                if (
                    len(pet.out_edges(child_cu_id, EdgeType.SUCCESSOR)) == 0
                    and len(pet.in_edges(child_cu_id, EdgeType.SUCCESSOR)) != 0
                ):
                    exit_cu_ids.add(child_cu_id)
        return exit_cu_ids

    def calculate_reachability_pairs(self, pet: PEGraphX):
        reachability_pairs: Dict[NodeID, Set[NodeID]] = dict()
        # create graph copy and remove all but successor edges
        copied_graph = pet.g.copy()

        # remove all but successor edges
        to_be_removed = set()
        for edge in copied_graph.edges:
            edge_data = cast(Dependency, copied_graph.edges[edge]["data"])
            if edge_data.etype != EdgeType.SUCCESSOR:
                to_be_removed.add(edge)
        for edge in to_be_removed:
            copied_graph.remove_edge(edge[0], edge[1])

        # calculate dfs successors for children CUs
        for node_id in cast(List[NodeID], self.children_cu_ids):
            reachability_pairs[node_id] = {node_id}
            successors = [t for s, t in nx.dfs_tree(copied_graph, node_id).edges()]
            reachability_pairs[node_id].update(successors)
        return reachability_pairs

    def get_immediate_post_dominators(self, pet: PEGraphX) -> Dict[NodeID, NodeID]:
        if self.immediate_post_dominators_present:
            import sys

            print("METADATA HIT! ", self.node_id, file=sys.stderr)
            return self.immediate_post_dominators
        else:
            # copy graph since edges need to be removed
            copied_graph = pet.g.copy()
            exit_cu_ids = self.get_exit_cu_ids(pet)
            # remove all but successor edges
            to_be_removed = set()
            for edge in copied_graph.edges:
                edge_data = cast(Dependency, copied_graph.edges[edge]["data"])
                if edge_data.etype != EdgeType.SUCCESSOR:
                    to_be_removed.add(edge)
            for edge in to_be_removed:
                copied_graph.remove_edge(edge[0], edge[1])

            # reverse edges
            immediate_post_dominators: Set[Tuple[NodeID, NodeID]] = set()
            for exit_cu_id in exit_cu_ids:
                immediate_post_dominators.update(nx.immediate_dominators(copied_graph.reverse(), exit_cu_id).items())

            immediate_post_dominators_dict = dict(immediate_post_dominators)
            # add trivial cases for missing modes
            for child_id in cast(List[NodeID], self.children_cu_ids):
                if child_id not in immediate_post_dominators_dict:
                    immediate_post_dominators_dict[child_id] = child_id

            # initialize result dictionary and add trivial cases for all children
            self.immediate_post_dominators = dict()
            for child_id in cast(List[NodeID], self.children_cu_ids):
                if child_id not in self.immediate_post_dominators:
                    self.immediate_post_dominators[child_id] = child_id

            # find post dominator outside parent, if type(parent) != function
            for node_id in cast(List[NodeID], self.children_cu_ids):
                if type(pet.node_at(node_id)) != CUNode:
                    continue
                # initialize search with immediate post dominator
                post_dom_id = immediate_post_dominators_dict[node_id]
                visited = set()
                use_original = False
                while (
                    pet.node_at(node_id).get_parent_id(pet) == pet.node_at(post_dom_id).get_parent_id(pet)
                    and type(pet.node_at(cast(NodeID, pet.node_at(post_dom_id).get_parent_id(pet)))) != FunctionNode
                ):
                    if post_dom_id in visited:
                        # cycle detected!
                        use_original = True
                        break

                    visited.add(post_dom_id)
                    new_post_dom_id = immediate_post_dominators_dict[post_dom_id]
                    import sys

                    print("Post dom: ", post_dom_id, file=sys.stderr)
                    print("New post dom: ", new_post_dom_id, file=sys.stderr)
                    print(file=sys.stderr)
                    if post_dom_id == new_post_dom_id:
                        break
                    post_dom_id = new_post_dom_id
                if not use_original:
                    # found post dom
                    self.immediate_post_dominators[node_id] = post_dom_id
            self.immediate_post_dominators_present = True
            return self.immediate_post_dominators

    def get_memory_accesses(
        self, writes_by_device: Dict[int, Dict[NodeID, Dict[MemoryRegion, Set[Optional[int]]]]]
    ) -> Dict[int, Dict[MemoryRegion, Set[Optional[int]]]]:
        if not self.memory_accesses_present:
            self.memory_accesses = dict()
            self.memory_accesses_present = True

        for child_id in cast(List[NodeID], self.children_cu_ids):
            for device_id in writes_by_device:
                if device_id not in self.memory_accesses:
                    self.memory_accesses[device_id] = dict()
                if child_id in writes_by_device[device_id]:
                    for mem_reg in writes_by_device[device_id][child_id]:
                        if mem_reg not in self.memory_accesses[device_id]:
                            self.memory_accesses[device_id][mem_reg] = set()
                        self.memory_accesses[device_id][mem_reg].update(writes_by_device[device_id][child_id][mem_reg])
        return self.memory_accesses


def parse_cu(node: ObjectifiedElement) -> Node:
    node_id = node.get("id")
    node_type = NodeType(int(node.get("type")))

    n: Node
    # CU Node
    if node_type == NodeType.CU:
        n = CUNode(node_id)
        if hasattr(node.localVariables, "local"):
            n.local_vars = [
                Variable(
                    v.get("type"),
                    v.text,
                    v.get("defLine"),
                    v.get("accessMode"),
                    int(v.get("sizeInByte")) if v.get("sizeInByte") is not None else 0,
                )
                for v in node.localVariables.local
            ]
        if hasattr(node.globalVariables, "global"):
            n.global_vars = [
                Variable(
                    v.get("type"),
                    v.text,
                    v.get("defLine"),
                    v.get("accessMode"),
                    int(v.get("sizeInByte")) if v.get("sizeInByte") is not None else 0,
                )
                for v in getattr(node.globalVariables, "global")
            ]
        if hasattr(node, "BasicBlockID"):
            n.basic_block_id = getattr(node, "BasicBlockID").text
        if hasattr(node, "returnInstructions"):
            n.return_instructions_count = int(getattr(node, "returnInstructions").get("count"))
        if hasattr(node.callsNode, "nodeCalled"):
            n.node_calls = [
                {"cuid": v.text, "atLine": v.get("atLine")}
                for v in getattr(node.callsNode, "nodeCalled")
                if v.get("atLine") is not None
            ]
        if hasattr(node, "callsNode") and hasattr(node.callsNode, "recursiveFunctionCall"):
            n.recursive_function_calls = [n.text for n in node.callsNode.recursiveFunctionCall]
        if hasattr(node, "performsFileIO"):
            n.performs_file_io = True if int(getattr(node, "performsFileIO")) == 1 else False
        n.instructions_count = int(getattr(node, "instructionsCount"))

    # FUNC or DUMMY NODE
    elif node_type == NodeType.DUMMY or node_type == NodeType.FUNC:
        dummy_or_func: Union[DummyNode, FunctionNode]
        if node_type == NodeType.DUMMY:
            dummy_or_func = DummyNode(node_id)
        else:
            dummy_or_func = FunctionNode(node_id)
        if hasattr(node, "funcArguments") and hasattr(node.funcArguments, "arg"):
            dummy_or_func.args = [
                Variable(
                    v.get("type"),
                    v.text,
                    v.get("defLine"),
                    sizeInByte=int(v.get("sizeInByte")) if v.get("sizeInByte") is not None else 0,
                )
                for v in node.funcArguments.arg
            ]
        n = dummy_or_func

    # LOOP Node
    elif node_type == NodeType.LOOP:
        n = LoopNode(node_id)
    else:
        assert False, "invalid NodeType"

    _, n.start_line = parse_id(node.get("startsAtLine"))
    _, n.end_line = parse_id(node.get("endsAtLine"))
    n.name = node.get("name")

    return n


def parse_dependency(dep: DependenceItem) -> Dependency:
    d = Dependency(EdgeType.DATA)
    d.source_line = dep.source
    d.sink_line = dep.sink
    # check for intra-iteration dependencies
    if "_II_" in dep.type:
        d.intra_iteration = True
        # get intra_iteration_level
        d.intra_iteration_level = int(dep.type[dep.type.rindex("_") + 1 :])
        d.dtype = DepType[dep.type[: dep.type.index("_")]]  # remove _II_x tag
    else:
        d.dtype = DepType[dep.type]
    d.var_name = dep.var_name
    d.memory_region = dep.memory_region
    return d


class PEGraphX(object):
    g: nx.MultiDiGraph
    reduction_vars: List[Dict[str, str]]
    main: Node
    pos: Dict[Any, Any]

    def __init__(self, g: nx.MultiDiGraph, reduction_vars: List[Dict[str, str]], pos):
        self.g = g
        self.reduction_vars = reduction_vars
        for _, node in g.nodes(data="data"):
            if node.name == "main":
                self.main = node
        self.pos = pos

    @classmethod
    def from_parsed_input(
        cls,
        cu_dict: Dict[str, ObjectifiedElement],
        dependencies_list: List[DependenceItem],
        loop_data: Dict[str, LoopData],
        reduction_vars: List[Dict[str, str]],
    ):
        """Constructor for making a PETGraphX from the output of parser.parse_inputs()"""
        g = nx.MultiDiGraph()
        print("\tCreating graph...")

        for id, node in cu_dict.items():
            n = parse_cu(node)
            g.add_node(id, data=n)

        print("\tAdded nodes...")

        for node_id, node in cu_dict.items():
            source = node_id
            if "successors" in dir(node) and "CU" in dir(node.successors):
                for successor in [n.text for n in node.successors.CU]:
                    if successor not in g:
                        print(f"WARNING: no successor node {successor} found")
                    # do not allow "self-successor" edges (incorrect, but not critical. might occur in Data.xml)
                    if source != successor:
                        g.add_edge(source, successor, data=Dependency(EdgeType.SUCCESSOR))

            if "callsNode" in dir(node) and "nodeCalled" in dir(node.callsNode):
                for nodeCalled in [n.text for n in node.callsNode.nodeCalled]:
                    if nodeCalled not in g:
                        print(f"WARNING: no nodeCalled {nodeCalled} found")
                    g.add_edge(source, nodeCalled, data=Dependency(EdgeType.CALLSNODE))

            if "childrenNodes" in dir(node):
                for child in [n.text for n in node.childrenNodes]:
                    if child not in g:
                        print(f"WARNING: no child node {child} found")
                    if not (source, child) in g.edges:
                        g.add_edge(source, child, data=Dependency(EdgeType.CHILD))

        print("\tAdded edges...")

        for _, node in g.nodes(data="data"):
            if isinstance(node, LoopNode):
                node.loop_data = loop_data.get(node.start_position(), None)
                # TODO remove loop_iterations property, was kept for backwards compatibility only
                if node.loop_data is not None:
                    # node.loop_iterations = node.loop_data.total_iteration_count
                    node.loop_iterations = node.loop_data.average_iteration_count

        print("\tAdded loop data...")

        # calculate position before dependencies affect them
        try:
            pos = nx.planar_layout(g)  # good
        except nx.exception.NetworkXException:
            try:
                # fallback layouts
                pos = nx.shell_layout(g)  # maybe
            except nx.exception.NetworkXException:
                pos = nx.random_layout(g)
        print("\tCalculated positions...")
        for idx, dep in enumerate(dependencies_list):
            if dep.type == "INIT":
                sink = readlineToCUIdMap[dep.sink]
                if len(sink) > 0:
                    for s in sink:
                        g.add_edge(s, s, data=parse_dependency(dep))
                else:
                    # check for write lines
                    sink = writelineToCUIdMap[dep.sink]
                    for s in sink:
                        g.add_edge(s, s, data=parse_dependency(dep))
                continue

            sink_cu_ids = readlineToCUIdMap[dep.sink]
            source_cu_ids = writelineToCUIdMap[dep.source]

            for idx_1, sink_cu_id in enumerate(sink_cu_ids):
                for idx_2, source_cu_id in enumerate(source_cu_ids):
                    sink_node = g.nodes[sink_cu_id]["data"]
                    source_node = g.nodes[source_cu_id]["data"]
                    vars_in_sink_node = set()
                    vars_in_source_node = set()
                    if type(sink_node) == CUNode:
                        for var in itertools.chain(sink_node.local_vars, sink_node.global_vars):
                            vars_in_sink_node.add(var.name)
                    if type(source_node) == CUNode:
                        for var in itertools.chain(source_node.local_vars, source_node.global_vars):
                            vars_in_source_node.add(var.name)

                    if dep.var_name not in vars_in_sink_node and dep.var_name not in vars_in_source_node:
                        continue
                    if sink_cu_id and source_cu_id:
                        g.add_edge(sink_cu_id, source_cu_id, data=parse_dependency(dep))
        print("\tAdded dependencies...")
        return cls(g, reduction_vars, pos)

    def map_static_and_dynamic_dependencies(self):
        print("\tMapping static to dynamic dependencies...")
        print("\t\tIdentifying mappings between static and dynamic memory regions...", end=" ")
        mem_reg_mappings: Dict[MemoryRegion, Set[MemoryRegion]] = dict()
        # initialize mappings
        for node_id in [n.id for n in self.all_nodes(CUNode)]:
            out_deps = [(s, t, d) for s, t, d in self.out_edges(node_id) if d.etype == EdgeType.DATA]

            # for outgoing dependencies, the scope must be equal
            # as a result, comparing variable names to match memory regions is valid
            for _, _, d1 in out_deps:
                for _, _, d2 in out_deps:
                    if d1 == d2:
                        continue
                    if d1.var_name == d2.var_name:
                        if d1.memory_region != d2.memory_region:
                            if d1.memory_region not in mem_reg_mappings:
                                mem_reg_mappings[d1.memory_region] = set()
                            if d2.memory_region not in mem_reg_mappings:
                                mem_reg_mappings[d2.memory_region] = set()
                            mem_reg_mappings[d1.memory_region].add(d2.memory_region)
                            mem_reg_mappings[d2.memory_region].add(d1.memory_region)
        print("Done.")

        print("\t\tInstantiating static dependencies...", end=" ")
        # create copies of static dependency edges for all dynamic mappings
        for node_id in [n.id for n in self.all_nodes(CUNode)]:
            out_deps = [(s, t, d) for s, t, d in self.out_edges(node_id) if d.etype == EdgeType.DATA]
            for s, t, d in out_deps:
                if d.memory_region.startswith("S"):
                    # Static dependency found
                    # check if mappings exist
                    if d.memory_region in mem_reg_mappings:
                        # create instances for all dynamic mappings
                        for dynamic_mapping in [
                            mapping for mapping in mem_reg_mappings[d.memory_region] if not mapping.startswith("S")
                        ]:
                            edge_data = copy.deepcopy(d)
                            edge_data.memory_region = dynamic_mapping
                            self.g.add_edge(s, t, data=edge_data)

        print("Done.")

    def calculateFunctionMetadata(self) -> None:
        # store id of parent function in each node
        # and store in each function node a list of all children ids
        func_nodes = self.all_nodes(FunctionNode)
        print("Calculating local metadata results for functions...")
        import tqdm  # type: ignore
        from multiprocessing import Pool
        from .parallel_utils import pet_function_metadata_initialize_worker, pet_function_metadata_parse_func

        param_list = func_nodes
        with Pool(initializer=pet_function_metadata_initialize_worker, initargs=(self,)) as pool:
            tmp_result = list(
                tqdm.tqdm(pool.imap_unordered(pet_function_metadata_parse_func, param_list), total=len(param_list))
            )
        # calculate global result
        print("Calculating global result...")
        global_reachability_dict: Dict[NodeID, Set[NodeID]] = dict()
        for local_result in tmp_result:
            parsed_function_id, local_reachability_dict, local_children_ids = local_result
            # set reachability values to function nodes
            cast(FunctionNode, self.node_at(parsed_function_id)).reachability_pairs = local_reachability_dict
            # set parent function for visited nodes
            for child_id in local_children_ids:
                self.node_at(child_id).parent_function_id = parsed_function_id

        print("\tMetadata calculation done.")

        # cleanup dependencies (remove dependencies, if it is overwritten by a more specific Intra-iteration dependency
        print("Cleaning duplicated dependencies...")
        to_be_removed = []
        for cu_node in self.all_nodes(CUNode):
            out_deps = self.out_edges(cu_node.id, EdgeType.DATA)
            for dep_1 in out_deps:
                for dep_2 in out_deps:
                    if dep_1 == dep_2:
                        continue
                    if (
                        dep_1[2].dtype == dep_2[2].dtype
                        and dep_1[2].etype == dep_2[2].etype
                        and dep_1[2].memory_region == dep_2[2].memory_region
                        and dep_1[2].sink_line == dep_2[2].sink_line
                        and dep_1[2].source_line == dep_2[2].source_line
                        and dep_1[2].var_name == dep_2[2].var_name
                    ):
                        if not dep_1[2].intra_iteration and dep_2[2].intra_iteration:
                            # dep_2 is a more specific duplicate of dep_1
                            # remove dep_1
                            to_be_removed.append(dep_1)

        to_be_removed_with_keys = []
        for dep in to_be_removed:
            graph_edges = self.g.out_edges(dep[0], keys=True, data="data")

            for s, t, key, data in graph_edges:
                if dep[0] == s and dep[1] == t and dep[2] == data:
                    to_be_removed_with_keys.append((s, t, key))
        for edge in set(to_be_removed_with_keys):
            self.g.remove_edge(edge[0], edge[1], edge[2])
        print("Cleaning dependencies done.")

        # cleanup dependencies II : only consider the Intra-iteration dependencies with the highest level
        print("Cleaning duplicated dependencies II...")
        to_be_removed = []
        for cu_node in self.all_nodes(CUNode):
            out_deps = self.out_edges(cu_node.id, EdgeType.DATA)
            for dep_1 in out_deps:
                for dep_2 in out_deps:
                    if dep_1 == dep_2:
                        continue
                    if (
                        dep_1[2].dtype == dep_2[2].dtype
                        and dep_1[2].etype == dep_2[2].etype
                        and dep_1[2].memory_region == dep_2[2].memory_region
                        and dep_1[2].sink_line == dep_2[2].sink_line
                        and dep_1[2].source_line == dep_2[2].source_line
                        and dep_1[2].var_name == dep_2[2].var_name
                        and dep_1[2].intra_iteration
                        and dep_2[2].intra_iteration
                    ):
                        if dep_1[2].intra_iteration_level < dep_2[2].intra_iteration_level:
                            # dep_2 originated from a deeper nesting level. Remove less specific duplicate dep_1.
                            to_be_removed.append(dep_1)

        to_be_removed_with_keys = []
        for dep in to_be_removed:
            graph_edges = self.g.out_edges(dep[0], keys=True, data="data")

            for s, t, key, data in graph_edges:
                if dep[0] == s and dep[1] == t and dep[2] == data:
                    to_be_removed_with_keys.append((s, t, key))
        for edge in set(to_be_removed_with_keys):
            self.g.remove_edge(edge[0], edge[1], edge[2])
        print("Cleaning dependencies II done.")

    def calculateLoopMetadata(self):
        print("Calculating loop metadata")

        # calculate loop indices
        print("Calculating loop indices")
        loop_nodes = self.all_nodes(LoopNode)
        for loop in loop_nodes:
            subtree = self.subtree_of_type(loop, CUNode)
            # get variables used in loop
            candidates: Set[Variable] = set()
            for node in subtree:
                candidates.update(node.global_vars + node.local_vars)
            # identify loop indices
            loop_indices: Set[Variable] = set()
            for v in candidates:
                if self.is_loop_index(v.name, loop.start_position(), subtree):
                    loop_indices.add(v)
            loop.loop_indices = [v.name for v in loop_indices]
        print("\tDone.")

        print("Calculating loop metadata done.")

    def show(self):
        """Plots the graph

        :return:
        """
        # print("showing")
        plt.plot()
        pos = self.pos

        # draw nodes
        nx.draw_networkx_nodes(
            self.g,
            pos=pos,
            node_size=200,
            node_color="#2B85FD",
            node_shape="o",
            nodelist=[n for n in self.g.nodes if isinstance(self.node_at(n), CUNode)],
        )
        nx.draw_networkx_nodes(
            self.g,
            pos=pos,
            node_size=200,
            node_color="#ff5151",
            node_shape="d",
            nodelist=[n for n in self.g.nodes if isinstance(self.node_at(n), LoopNode)],
        )
        nx.draw_networkx_nodes(
            self.g,
            pos=pos,
            node_size=200,
            node_color="grey",
            node_shape="s",
            nodelist=[n for n in self.g.nodes if isinstance(self.node_at(n), DummyNode)],
        )
        nx.draw_networkx_nodes(
            self.g,
            pos=pos,
            node_size=200,
            node_color="#cf65ff",
            node_shape="s",
            nodelist=[n for n in self.g.nodes if isinstance(self.node_at(n), FunctionNode)],
        )
        nx.draw_networkx_nodes(
            self.g,
            pos=pos,
            node_color="yellow",
            node_shape="h",
            node_size=250,
            nodelist=[n for n in self.g.nodes if self.node_at(n).name == "main"],
        )
        # id as label
        labels = {}
        for n in self.g.nodes:
            labels[n] = str(self.g.nodes[n]["data"])
        nx.draw_networkx_labels(self.g, pos, labels, font_size=7)

        nx.draw_networkx_edges(
            self.g,
            pos,
            edgelist=[e for e in self.g.edges(data="data") if e[2].etype == EdgeType.CHILD],
        )
        nx.draw_networkx_edges(
            self.g,
            pos,
            edge_color="green",
            edgelist=[e for e in self.g.edges(data="data") if e[2].etype == EdgeType.SUCCESSOR],
        )
        nx.draw_networkx_edges(
            self.g,
            pos,
            edge_color="red",
            edgelist=[e for e in self.g.edges(data="data") if e[2].etype == EdgeType.DATA],
        )
        nx.draw_networkx_edges(
            self.g,
            pos,
            edge_color="yellow",
            edgelist=[e for e in self.g.edges(data="data") if e[2].etype == EdgeType.CALLSNODE],
        )
        nx.draw_networkx_edges(
            self.g,
            pos,
            edge_color="orange",
            edgelist=[e for e in self.g.edges(data="data") if e[2].etype == EdgeType.PRODUCE_CONSUME],
        )

        plt.show()
        plt.savefig("graphX.svg")

    def node_at(self, node_id: NodeID) -> Node:
        """Gets node data by node id

        :param node_id: id of the node
        :return: Node
        """
        return cast(Node, self.g.nodes[node_id]["data"])

    # generic type for subclasses of Node
    NodeT = TypeVar("NodeT", bound=Node)

    @overload
    def all_nodes(self) -> List[Node]:
        ...

    @overload
    def all_nodes(self, type: Union[Type[NodeT], Tuple[Type[NodeT], ...]]) -> List[NodeT]:
        ...

    def all_nodes(self, type=Node):
        """List of all nodes of specified type

        :param type: type(s) of nodes
        :return: List of all nodes
        """
        return [n[1] for n in self.g.nodes(data="data") if isinstance(n[1], type)]

    def out_edges(
        self, node_id: NodeID, etype: Optional[Union[EdgeType, List[EdgeType]]] = None
    ) -> List[Tuple[NodeID, NodeID, Dependency]]:
        """Get outgoing edges of node of specified type

        :param node_id: id of the source node
        :param etype: type of edges
        :return: list of outgoing edges
        """
        if etype is None:
            return [t for t in self.g.out_edges(node_id, data="data")]
        elif type(etype) == list:
            return [t for t in self.g.out_edges(node_id, data="data") if t[2].etype in etype]
        else:
            return [t for t in self.g.out_edges(node_id, data="data") if t[2].etype == etype]

    def in_edges(
        self, node_id: NodeID, etype: Optional[Union[EdgeType, List[EdgeType]]] = None
    ) -> List[Tuple[NodeID, NodeID, Dependency]]:
        """Get incoming edges of node of specified type

        :param node_id: id of the target node
        :param etype: type of edges
        :return: list of incoming edges
        """
        if etype is None:
            return [t for t in self.g.in_edges(node_id, data="data")]
        elif type(etype) == list:
            return [t for t in self.g.in_edges(node_id, data="data") if t[2].etype in etype]
        else:
            return [t for t in self.g.in_edges(node_id, data="data") if t[2].etype == etype]

    @overload
    def subtree_of_type(self, root: Node) -> List[Node]:
        ...

    @overload
    def subtree_of_type(self, root: Node, type: Union[Type[NodeT], Tuple[Type[NodeT], ...]]) -> List[NodeT]:
        ...

    def subtree_of_type(self, root, type=Node):
        """Gets all nodes in subtree of specified type including root

        :param root: root node
        :param type: type of children, None is equal to a wildcard
        :return: list of nodes in subtree
        """
        return self.subtree_of_type_rec(root, set(), type)

    @overload
    def subtree_of_type_rec(self, root: Node, visited: Set[Node]) -> List[Node]:
        ...

    @overload
    def subtree_of_type_rec(
        self, root: Node, visited: Set[Node], type: Union[Type[NodeT], Tuple[Type[NodeT], ...]]
    ) -> List[NodeT]:
        ...

    def subtree_of_type_rec(self, root, visited, type=Node):
        """recursive helper function for subtree_of_type"""
        # check if root is of type target
        res = []
        if isinstance(root, type):
            res.append(root)

        # append root to visited
        visited.add(root)

        # enter recursion
        for _, target, _ in self.out_edges(root.id, [EdgeType.CHILD, EdgeType.CALLSNODE]):
            # prevent cycles
            if self.node_at(target) in visited:
                continue
            res += self.subtree_of_type_rec(self.node_at(target), visited, type)

        return res

    def __cu_equal__(self, cu_1: Node, cu_2: Node):
        """Alternative to CUNode.__eq__, bypasses the isinstance-check and relies on MyPy for type safety.
        :param cu_1: CUNode 1
        :param cu_2: CUNode 2
        :return: True, if cu_1 == cu_2. False, else"""
        if cu_1.file_id == cu_2.file_id and cu_1.node_id == cu_2.node_id:
            return True
        return False

    def direct_successors(self, root: Node) -> List[Node]:
        """Gets only direct successors of any type

        :param root: root node
        :return: list of direct successors
        """
        return [self.node_at(t) for s, t, d in self.out_edges(root.id, EdgeType.SUCCESSOR)]

    def direct_children_or_called_nodes(self, root: Node) -> List[Node]:
        """Gets direct children of any type. Also includes nodes of called functions

        :param root: root node
        :return: list of direct children
        """
        return [self.node_at(t) for s, t, d in self.out_edges(root.id, [EdgeType.CHILD, EdgeType.CALLSNODE])]

    def direct_children(self, root: Node) -> List[Node]:
        """Gets direct children of any type. This includes called nodes!

        :param root: root node
        :return: list of direct children
        """
        return [self.node_at(t) for s, t, d in self.out_edges(root.id, EdgeType.CHILD)]

    def direct_children_or_called_nodes_of_type(self, root: Node, type: Type[NodeT]) -> List[NodeT]:
        """Gets only direct children of specified type. This includes called nodes!

        :param root: root node
        :param type: type of children
        :return: list of direct children
        """
        nodes = [self.node_at(t) for s, t, d in self.out_edges(root.id, [EdgeType.CHILD, EdgeType.CALLSNODE])]

        return [t for t in nodes if isinstance(t, type)]

    def is_reduction_var(self, line: LineID, name: str) -> bool:
        """Determines, whether or not the given variable is reduction variable

        :param line: loop line number
        :param name: variable name
        :return: true if is reduction variable
        """
        return any(rv for rv in self.reduction_vars if rv["loop_line"] == line and rv["name"] == name)

    def depends_ignore_readonly(self, source: Node, target: Node, root_loop: Node) -> bool:
        """Detects if source node or one of it's children has a RAW dependency to target node or one of it's children
        The loop index and readonly variables are ignored.

        :param source: source node for dependency detection (later occurrence in the source code)
        :param target: target of dependency (prior occurrence in the source code)
        :param root_loop: root loop
        :return: true, if there is RAW dependency"""
        if source == target:
            return False

        # get recursive children of source and target
        source_children_ids = [node.id for node in self.subtree_of_type(source, CUNode)]
        target_children_ids = [node.id for node in self.subtree_of_type(target, CUNode)]

        # get required metadata
        loop_start_lines: List[LineID] = []
        root_children = self.subtree_of_type(root_loop, (CUNode, LoopNode))
        root_children_cus = [cu for cu in root_children if isinstance(cu, CUNode)]
        root_children_loops = [cu for cu in root_children if isinstance(cu, LoopNode)]
        for v in root_children_loops:
            loop_start_lines.append(v.start_position())

        # check for RAW dependencies between any of source_children and any of target_children
        for source_child_id in source_children_ids:
            # get a list of filtered dependencies, outgoing from source_child
            out_deps = self.out_edges(source_child_id, EdgeType.DATA)
            out_raw_deps = [dep for dep in out_deps if dep[2].dtype == DepType.RAW]
            filtered_deps = [
                elem
                for elem in out_raw_deps
                if not self.is_readonly_inside_loop_body(
                    elem[2],
                    root_loop,
                    root_children_cus,
                    root_children_loops,
                    loops_start_lines=loop_start_lines,
                )
            ]
            filtered_deps = [
                elem
                for elem in filtered_deps
                if not self.is_loop_index(elem[2].var_name, loop_start_lines, root_children_cus)
            ]
            # get a list of dependency targets
            dep_targets = [t for _, t, _ in filtered_deps]
            # check if overlap between dependency targets and target_children exists.
            overlap = [node_id for node_id in dep_targets if node_id in target_children_ids]
            if len(overlap) > 0:
                # if so, a RAW dependency exists
                return True
        return False

    def unused_check_alias(self, s: NodeID, t: NodeID, d: Dependency, root_loop: Node) -> bool:
        sub = self.subtree_of_type(root_loop, CUNode)
        parent_func_sink = self.get_parent_function(self.node_at(s))
        parent_func_source = self.get_parent_function(self.node_at(t))

        res = False
        d_var_name_str = cast(str, d.var_name)

        if self.unused_is_global(d_var_name_str, sub) and not (
            self.is_passed_by_reference(d, parent_func_sink) and self.is_passed_by_reference(d, parent_func_source)
        ):
            return res
        return not res

    def unused_is_global(self, var: str, tree: Sequence[Node]) -> bool:
        """Checks if variable is global

        :param var: variable name
        :param tree: nodes to search
        :return: true if global
        """

        for node in tree:
            if isinstance(node, CUNode):
                for gv in node.global_vars:
                    if gv.name == var:
                        # TODO from tmp global vars
                        return False
        return False

    def is_passed_by_reference(self, dep: Dependency, func: FunctionNode) -> bool:
        res = False

        for i in func.args:
            if i.name == dep.var_name:
                return True
        return False

    def unused_get_first_written_vars_in_loop(
        self, undefinedVarsInLoop: List[Variable], node: Node, root_loop: Node
    ) -> Set[Variable]:
        root_children = self.subtree_of_type(root_loop, CUNode)
        loop_node_ids = [n.id for n in root_children]
        fwVars = set()

        raw = set()
        war = set()
        waw = set()
        sub = root_children
        for sub_node in sub:
            raw.update(self.get_dep(sub_node, DepType.RAW, False))
            war.update(self.get_dep(sub_node, DepType.WAR, False))
            waw.update(self.get_dep(sub_node, DepType.WAW, False))

        for var in undefinedVarsInLoop:
            if var not in fwVars:
                for i in raw:
                    if i[2].var_name == var and i[0] in loop_node_ids and i[1] in loop_node_ids:
                        for e in itertools.chain(war, waw):
                            if e[2].var_name == var and e[0] in loop_node_ids and e[1] in loop_node_ids:
                                if e[2].sink_line == i[2].source_line:
                                    fwVars.add(var)

        return fwVars

    def get_dep(self, node: Node, dep_type: DepType, reversed: bool) -> List[Tuple[NodeID, NodeID, Dependency]]:
        """Searches all dependencies of specified type

        :param node: node
        :param dep_type: type of dependency
        :param reversed: if true the it looks for incoming dependencies
        :return: list of dependencies
        """
        return [
            e
            for e in (self.in_edges(node.id, EdgeType.DATA) if reversed else self.out_edges(node.id, EdgeType.DATA))
            if e[2].dtype == dep_type
        ]

    def is_scalar_val(self, allVars: List[Variable], var: str) -> bool:
        """Checks if variable is a scalar value

        :param var: variable
        :return: true if scalar
        """
        for x in allVars:
            if x.name == var:
                return not (x.type.endswith("**") or x.type.startswith("ARRAY" or x.type.startswith("[")))
            else:
                return False
        raise ValueError("allVars must not be empty.")

    def get_variables(self, nodes: Sequence[Node]) -> Dict[Variable, Set[MemoryRegion]]:
        """Gets all variables and corresponding memory regions in nodes

        :param nodes: nodes
        :return: Set of variables
        """
        res: Dict[Variable, Set[MemoryRegion]] = dict()
        for node in nodes:
            if isinstance(node, CUNode):
                for v in node.local_vars:
                    if v not in res:
                        res[v] = set()
                for v in node.global_vars:
                    if v not in res:
                        res[v] = set()
                # try to identify memory regions
                for var_name in res:
                    # since the variable name is checked for equality afterwards,
                    # it is safe to consider incoming dependencies at this point as well.
                    # Note that INIT type edges are considered as well!
                    for _, _, dep in self.out_edges(node.id, EdgeType.DATA) + self.in_edges(node.id, EdgeType.DATA):
                        if dep.var_name == var_name.name:
                            if dep.memory_region is not None:
                                res[var_name].add(dep.memory_region)
        return res

    def get_undefined_variables_inside_loop(
        self, root_loop: Node, include_global_vars: bool = False
    ) -> Dict[Variable, Set[MemoryRegion]]:
        sub = self.subtree_of_type(root_loop, CUNode)
        vars = self.get_variables(sub)
        dummyVariables = []
        definedVarsInLoop = []
        definedVarsInCalledFunctions = []

        # Remove llvm temporary variables
        for var in vars:
            if var.defLine == "LineNotFound" or "0:" in var.defLine:
                dummyVariables.append(var)
            elif not include_global_vars:
                if var.defLine == "GlobalVar" and not self.is_reduction_var(root_loop.start_position(), var.name):
                    dummyVariables.append(var)

        # vars = list(set(vars) ^ set(dummyVariables))
        for key in set(dummyVariables):
            if key in vars:
                del vars[key]

        # Exclude variables which are defined inside the loop
        for var in vars:
            if var.defLine >= root_loop.start_position() and var.defLine <= root_loop.end_position():
                definedVarsInLoop.append(var)

        # vars = list(set(vars) ^ set(definedVarsInLoop))
        for key in set(definedVarsInLoop):
            if key in vars:
                del vars[key]

        # Also, exclude variables which are defined inside
        # functions that are called within the loop
        for var in vars:
            for s in sub:
                if var.defLine >= s.start_position() and var.defLine <= s.end_position():
                    definedVarsInCalledFunctions.append(var)

        # vars = list(set(vars) ^ set(definedVarsInCalledFunctions))
        for key in set(definedVarsInCalledFunctions):
            if key in vars:
                del vars[key]

        return vars

    def unused_is_first_written_in_loop(self, dep: Dependency, root_loop: Node):
        """Checks whether a variable is first written inside the current node

        :param var:
        :param raw_deps: raw dependencies of the loop
        :param war_deps: war dependencies of the loop
        :param reverse_raw_deps:
        :param reverse_war_deps:
        :param tree: subtree of the loop
        :return: true if first written
        """

        result = False
        children = self.subtree_of_type(root_loop, CUNode)

        for v in children:
            for t, d in [
                (t, d)
                for s, t, d in self.out_edges(v.id, EdgeType.DATA)
                if d.dtype == DepType.WAR or d.dtype == DepType.WAW
            ]:
                if d.var_name is None:
                    return False
                assert d.var_name is not None
                if dep.var_name == d.var_name:
                    if dep.source_line == d.sink_line:
                        result = True
                        break
                # None may occur because __get_variables doesn't check for actual elements
        return result

    def is_loop_index(self, var_name: Optional[str], loops_start_lines: List[LineID], children: Sequence[Node]) -> bool:
        """Checks, whether the variable is a loop index.

        :param var_name: name of the variable
        :param loops_start_lines: start lines of the loops
        :param children: children nodes of the loops
        :return: true if edge represents loop index
        """

        # If there is a raw dependency for var, the source cu is part of the loop
        # and the dependency occurs in loop header, then var is loop index+

        for c in children:
            for t, d in [
                (t, d)
                for s, t, d in self.out_edges(c.id, EdgeType.DATA)
                if d.dtype == DepType.RAW and d.var_name == var_name
            ]:
                if d.sink_line == d.source_line and d.source_line in loops_start_lines and self.node_at(t) in children:
                    return True

        return False

    def is_readonly_inside_loop_body(
        self,
        dep: Dependency,
        root_loop: Node,
        children_cus: Sequence[Node],
        children_loops: Sequence[Node],
        loops_start_lines: Optional[List[LineID]] = None,
    ) -> bool:
        """Checks, whether a variable is read-only in loop body

        :param dep: dependency variable
        :param root_loop: root loop
        :return: true if variable is read-only in loop body
        """
        if loops_start_lines is None:
            loops_start_lines = [v.start_position() for v in children_loops]

        for v in children_cus:
            for t, d in [
                (t, d)
                for s, t, d in self.out_edges(v.id, EdgeType.DATA)
                if d.dtype == DepType.WAR or d.dtype == DepType.WAW
            ]:
                # If there is a waw dependency for var, then var is written in loop
                # (sink is always inside loop for waw/war)
                if dep.memory_region == d.memory_region and not (d.sink_line in loops_start_lines):
                    return False
            for t, d in [(t, d) for s, t, d in self.in_edges(v.id, EdgeType.DATA) if d.dtype == DepType.RAW]:
                # If there is a reverse raw dependency for var, then var is written in loop
                # (source is always inside loop for reverse raw)
                if dep.memory_region == d.memory_region and not (d.source_line in loops_start_lines):
                    return False
        return True

    def get_parent_function(self, node: Node) -> FunctionNode:
        """Finds the parent of a node

        :param node: current node
        :return: node of parent function
        """
        if isinstance(node, FunctionNode):
            return node
        assert node.parent_function_id
        return cast(FunctionNode, self.node_at(node.parent_function_id))

    def get_left_right_subtree(
        self, target: Node, right_subtree: bool, ignore_called_nodes: bool = False
    ) -> List[Node]:
        """Searches for all subnodes of main which are to the left or to the right of the specified node

        :param target: node that divides the tree
        :param right_subtree: true - right subtree, false - left subtree
        :return: list of nodes in the subtree
        """
        stack: List[Node] = []
        res: List[Node] = []
        visited = set()

        parent_func = self.get_parent_function(target)
        stack.append(parent_func)

        while stack:
            current = stack.pop()

            if current == target:
                return res
            if isinstance(current, CUNode):
                res.append(current)

            if current in visited:  # suppress looping
                continue
            else:
                visited.add(current)

            if not ignore_called_nodes:
                stack.extend(
                    self.direct_children_or_called_nodes(current)
                    if right_subtree
                    else reversed(self.direct_children_or_called_nodes(current))
                )
            else:
                stack.extend(
                    self.direct_children(current) if right_subtree else reversed(self.direct_children(current))
                )
        return res

    def path(self, source: Node, target: Node) -> List[Node]:
        """DFS from source to target over edges of child type

        :param source: source node
        :param target: target node
        :return: list of nodes from source to target
        """
        return self.__path_rec(source, target, set())

    def __path_rec(self, source: Node, target: Node, visited: Set[Node]) -> List[Node]:
        """DFS from source to target over edges of child type

        :param source: source node
        :param target: target node
        :return: list of nodes from source to target
        """
        visited.add(source)
        if source == target:
            return [source]

        for child in [c for c in self.direct_children_or_called_nodes(source) if c not in visited]:
            path = self.__path_rec(child, target, visited)
            if path:
                path.insert(0, source)
                return path
        return []

    def get_reduction_sign(self, line: str, name: str) -> str:
        """Returns reduction operation for variable

        :param line: loop line number
        :param name: variable name
        :return: reduction operation
        """
        for rv in self.reduction_vars:
            if rv["loop_line"] == line and rv["name"] == name:
                return rv["operation"]
        return ""

    def dump_to_pickled_json(self):
        """Encodes and returns the entire Object into a pickled json string.
        The encoded string can be reconstructed into an object by using:
        jsonpickle.decode(json_str)

        :return: encoded string
        """
        return jsonpickle.encode(self)

    def check_reachability(self, target: Node, source: Node, edge_types: List[EdgeType]) -> bool:
        """check if target is reachable from source via edges of types edge_type.
        :param pet: PET graph
        :param source: CUNode
        :param target: CUNode
        :param edge_types: List[EdgeType]
        :return: Boolean"""
        if source == target:
            return True
        visited: List[str] = []
        queue = [target]
        while len(queue) > 0:
            cur_node = queue.pop(0)
            visited.append(cur_node.id)
            tmp_list = [
                (s, t, e) for s, t, e in self.in_edges(cur_node.id) if s not in visited and e.etype in edge_types
            ]
            for e in tmp_list:
                if self.node_at(e[0]) == source:
                    return True
                else:
                    if e[0] not in visited:
                        queue.append(self.node_at(e[0]))
        return False

    def is_predecessor(self, source_id: NodeID, target_id: NodeID) -> bool:
        """returns true, if source is a predecessor of target.
        This analysis includes traversal of successor, child and calls edges."""
        # if source and target_id are located within differenct functions, consider the callees instead of source_id
        source_parent_function = self.get_parent_function(self.node_at(source_id))
        target_parent_function = self.get_parent_function(self.node_at(target_id))
        if source_parent_function != target_parent_function:
            for callee_id in [s for s, _, _ in self.in_edges(source_parent_function.id, EdgeType.CALLSNODE)]:
                if self.is_predecessor(callee_id, target_id):
                    return True

        # if target is a loop node, get the first child of the loop, i.e. the entry node into the loop
        target_node = self.node_at(target_id)
        if target_node.type == NodeType.LOOP:
            target_id = self.direct_children(target_node)[0].id

        # perform a bfs search for target
        queue: List[NodeID] = [source_id]
        visited: List[NodeID] = []
        while queue:
            current = queue.pop(0)
            if current == target_id:
                return True
            visited.append(current)
            # add direct successors to queue
            queue += [
                n.id for n in self.direct_successors(self.node_at(current)) if n.id not in visited and n.id not in queue
            ]
            # add children to queue
            queue += [
                n.id for n in self.direct_children(self.node_at(current)) if n.id not in visited and n.id not in queue
            ]
            # add called functions to queue
            queue += [
                t for _, t, _ in self.out_edges(current, EdgeType.CALLSNODE) if t not in visited and t not in queue
            ]
        return False

    def check_reachability_and_get_path_nodes(
        self, target: CUNode, source: CUNode, edge_types: List[EdgeType]
    ) -> Tuple[bool, List[CUNode]]:
        """check if target is reachable from source via edges of types edge_type.
        :param pet: PET graph
        :param source: CUNode
        :param target: CUNode
        :param edge_types: List[EdgeType]
        :return: Boolean"""
        if source == target:
            return True, []

        # trivially not reachable
        if (
            self.get_parent_function(target) != self.get_parent_function(source)
            and EdgeType.CALLSNODE not in edge_types
        ):
            print("TRIVIAL FALSE!: ", source, target)
            return False, []

        visited: List[NodeID] = []
        queue: List[Tuple[CUNode, List[CUNode]]] = [(target, [])]
        while len(queue) > 0:
            cur_node, cur_path = queue.pop(0)
            visited.append(cur_node.id)
            tmp_list = [
                (s, t, e) for s, t, e in self.in_edges(cur_node.id) if s not in visited and e.etype in edge_types
            ]
            for e in tmp_list:
                if self.node_at(e[0]) == source:
                    return True, cur_path
                else:
                    if e[0] not in visited:
                        tmp_path = copy.deepcopy(cur_path)
                        tmp_path.append(cur_node)
                        queue.append((cast(CUNode, self.node_at(e[0])), tmp_path))
        return False, []

    def dump_to_gephi_file(self, name="pet.gexf"):
        """Note: Destroys the PETGraph!"""
        # replace node data with label
        for node_id in self.g.nodes:
            tmp_cu = self.g.nodes[node_id]["data"]
            del self.g.nodes[node_id]["data"]
            self.g.nodes[node_id]["id"] = tmp_cu.id
            self.g.nodes[node_id]["type"] = str(tmp_cu.type)
        for edge in self.g.edges:
            dep: Dependency = self.g.edges[edge]["data"]
            del self.g.edges[edge]["data"]
            self.g.edges[edge]["edge_type"] = str(dep.etype.name)
            if dep.etype == EdgeType.DATA:
                self.g.edges[edge]["var"] = dep.var_name
                self.g.edges[edge]["dep_type"] = str(dep.dtype.name)
        nx.write_gexf(self.g, name)

    def get_variable(self, root_node_id: NodeID, var_name: str) -> Optional[Variable]:
        """Search for the type of the given variable by BFS searching through successor edges in reverse, starting from
        the given root node, and checking the global and local vars of each encountered CU node."""
        queue: List[NodeID] = [root_node_id]
        visited: Set[NodeID] = set()
        while queue:
            current = queue.pop(0)
            current_node = cast(CUNode, self.node_at(current))
            visited.add(current)
            variables = current_node.local_vars + current_node.global_vars
            for v in variables:
                if v.name == var_name:
                    return v
            # add predecessors of current to the list
            predecessors = [s for s, t, d in self.in_edges(current, EdgeType.SUCCESSOR)]
            for pred in predecessors:
                if pred not in visited and pred not in queue:
                    queue.append(pred)
        return None

    def get_memory_regions(self, nodes: List[CUNode], var_name: str) -> Set[MemoryRegion]:
        """check dependencies of nodes for usages of 'var_name' and extract memory regions related to this name"""
        mem_regs: Set[MemoryRegion] = set()
        for node in nodes:
            out_deps = self.out_edges(node.id, EdgeType.DATA)
            for s, t, d in out_deps:
                if d.var_name == var_name:
                    if d.memory_region is not None:
                        mem_regs.add(d.memory_region)
        return mem_regs

    def get_path_nodes_between(self, target: CUNode, source: CUNode, edge_types: List[EdgeType]) -> List[CUNode]:
        """get all nodes of all patch which allow reaching target from source via edges of types edge_type.
        :param pet: PET graph
        :param source: CUNode
        :param target: CUNode
        :param edge_types: List[EdgeType]
        :return: List of encountered nodes"""

        visited: List[NodeID] = []
        queue: List[Tuple[CUNode, List[CUNode]]] = [
            (cast(CUNode, self.node_at(t)), [])
            for s, t, d in self.out_edges(source.id, edge_types)
            if type(self.node_at(t)) == CUNode
        ]

        while len(queue) > 0:
            cur_node, cur_path = queue.pop(0)
            visited.append(cur_node.id)
            tmp_list = [
                (s, t, e) for s, t, e in self.out_edges(cur_node.id) if t not in visited and e.etype in edge_types
            ]
            for e in tmp_list:
                if self.node_at(e[1]) == target or self.node_at(e[1]) == source:
                    continue
                else:
                    if e[1] not in visited:
                        tmp_path = copy.deepcopy(cur_path)
                        tmp_path.append(cur_node)
                        queue.append((cast(CUNode, self.node_at(e[1])), tmp_path))
        return [cast(CUNode, self.node_at(nid)) for nid in set(visited)]



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
import os.path
import shutil
from typing import Dict, List, Tuple, cast
import warnings

import jsonpickle  # type: ignore
from sympy import Float, Symbol  # type: ignore

from discopop_library.CodeGenerator.CodeGenerator import from_json_strings
from discopop_library.JSONHandler.JSONHandler import read_patterns_from_json_to_json
from discopop_library.PatchGenerator.PatchGeneratorArguments import PatchGeneratorArguments
from discopop_library.PatchGenerator.diffs import get_diffs_from_modified_code
from discopop_library.PathManagement.PathManagement import load_file_mapping
from discopop_library.discopop_optimizer.CostModels.DataTransfer.DataTransferCosts import add_data_transfer_costs
from discopop_library.discopop_optimizer.CostModels.utilities import get_performance_models_for_functions
from discopop_library.discopop_optimizer.DataTransfers.DataTransfers import calculate_data_transfers
from discopop_library.discopop_optimizer.OptimizerArguments import OptimizerArguments
from discopop_library.discopop_optimizer.PETParser.PETParser import PETParser
from discopop_library.discopop_optimizer.UpdateOptimization.main import optimize_updates
from discopop_library.discopop_optimizer.Variables.ExperimentUtils import (
    create_optimization_graph,
    export_detection_result_to_json,
    export_patterns_to_json,
    export_to_json,
    get_sequential_cost_model,
    initialize_free_symbol_ranges_and_distributions,
)
from discopop_library.discopop_optimizer.classes.enums.Distributions import FreeSymbolDistribution
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.classes.system.system_utils import generate_default_system_configuration
from discopop_library.discopop_optimizer.gui.queries.ValueTableQuery import query_user_for_symbol_values
from discopop_library.discopop_optimizer.optimization.evaluate import evaluate_configuration
from discopop_library.discopop_optimizer.optimization.evaluate_all_decision_combinations import (
    evaluate_all_decision_combinations,
)
from discopop_library.discopop_optimizer.optimization.evolutionary_algorithm import perform_evolutionary_search
from discopop_library.discopop_optimizer.suggestions.optimizers.main import optimize_suggestions
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at
from discopop_library.discopop_optimizer.utilities.visualization.update_graph import show_update_graph
from discopop_library.global_data.version.utils import get_version
from discopop_library.result_classes.DetectionResult import DetectionResult
from discopop_library.discopop_optimizer.classes.system.System import System
from discopop_library.discopop_optimizer.Microbench.ExtrapInterpolatedMicrobench import (
    ExtrapInterpolatedMicrobench,
)
from discopop_library.discopop_optimizer.Microbench.Microbench import MicrobenchType
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.utilities.MOGUtilities import (
    get_available_decisions_for_functions,
    show,
)
from discopop_library.discopop_optimizer.suggestions.importers.main import import_suggestions


def run(arguments: OptimizerArguments):
    # check prerequisites and setup folder structure
    if arguments.verbose:
        print("Started DiscoPoP Optimizer...")
        print("Creating optimizer directory...")
    optimizer_dir = os.path.join(os.getcwd(), "optimizer")
    if not os.path.exists(optimizer_dir):
        os.mkdir(optimizer_dir)

    explorer_dir = os.path.join(os.getcwd(), "explorer")
    profiler_dir = os.path.join(os.getcwd(), "profiler")
    pattern_file_path = os.path.join(explorer_dir, "patterns.json")
    if not os.path.exists(pattern_file_path):
        raise FileNotFoundError(
            "No pattern file found. Please execute the discopop_explorer in advance."
            + "\nExpected pattern file: "
            + pattern_file_path
        )
    file_mapping_path = os.path.join(os.getcwd(), "FileMapping.txt")
    if not os.path.exists(file_mapping_path):
        raise FileNotFoundError(
            "No file mapping found. Please execute the discopop_explorer in advance."
            + "\nExpected file: "
            + file_mapping_path
        )
    if not os.path.exists(arguments.system_configuration_path):
        generate_default_system_configuration(arguments.system_configuration_path)

    # create a new session, load data from previous steps)
    if arguments.verbose:
        print("Loading file mapping...", end="")
    file_mapping = load_file_mapping(file_mapping_path)
    if arguments.verbose:
        print("Done.")

    if arguments.verbose:
        print("Loading patterns...", end="")
    patterns_by_type = read_patterns_from_json_to_json(pattern_file_path, [])
    if arguments.verbose:
        print("Done.")

    if arguments.verbose:
        print("Loading detection result and PET...", end="")
    detection_result_dump_str = ""
    with open(os.path.join(explorer_dir, "detection_result_dump.json")) as f:
        detection_result_dump_str = f.read()
    detection_result: DetectionResult = jsonpickle.decode(detection_result_dump_str)

    if "version" in detection_result.__dict__:  # for backwards compatibility
        if detection_result.version != get_version():
            warnings.warn(
                "Restored DetectionResult was created using different DiscoPoP version: "
                + detection_result.version
                + "; current: "
                + get_version()
            )
    else:
        warnings.warn("Restored DetectionResult does not contain version information and might be outdated!")

    if arguments.verbose:
        print("Done")

    # define System
    # todo make system user-configurable, or detect it using a set of benchmarks
    system = System(arguments)

    # todo connections between devices might happen as update to host + update to second device.
    #  As of right now, connections between two devices are implemented in this manner.
    # todo check if OpenMP allows direct data transfers between devices

    # load overhead measurements into system if existent
    if arguments.doall_microbench_file != "None":
        # construct and set overhead model for doall suggestions
        system.set_device_doall_overhead_model(
            system.get_device(system.get_host_device_id()),
            ExtrapInterpolatedMicrobench(arguments.doall_microbench_file).getFunctionSympy(),
            arguments,
        )
    if arguments.reduction_microbench_file != "None":
        # construct and set overhead model for reduction suggestions
        system.set_reduction_overhead_model(
            system.get_device(system.get_host_device_id()),
            ExtrapInterpolatedMicrobench(arguments.reduction_microbench_file).getFunctionSympy(
                benchType=MicrobenchType.FOR
            ),
            arguments,
        )

    # define Experiment
    experiment = Experiment(file_mapping, system, detection_result, profiler_dir, arguments)

    # build optimization graph
    if arguments.verbose:
        print("Creating optimization graph...")
    create_optimization_graph(experiment, arguments)

    if arguments.verbose:
        print("Done.")
    # import parallelization suggestions
    experiment.optimization_graph = import_suggestions(experiment)
    # optimize parallelization suggestions
    experiment.optimization_graph = optimize_suggestions(experiment)

    if arguments.plot:
        show(experiment.optimization_graph, show_dataflow=False, show_mutex_edges=False)

    if arguments.verbose:
        print("# SUGGESTION ID -> NODE ID MAPPING")
        for suggestion_id in experiment.suggestion_to_node_ids_dict:
            print("#", suggestion_id, "->", experiment.suggestion_to_node_ids_dict[suggestion_id])
        print()

    # safety precaution: make sure the correct node ids are used everywhere
    for node_id in experiment.optimization_graph.nodes:
        node_data = data_at(experiment.optimization_graph, node_id)
        if node_id != node_data.node_id:
            node_data.node_id = node_id

    # get values for free symbols
    initialize_free_symbol_ranges_and_distributions(experiment, arguments, system)

    if arguments.verbose:
        print("# SUBSTITUTIONS:")
        for key in experiment.substitutions:
            print("#", key, " ->", experiment.substitutions[key])
        print()

    # calculate options for easy access
    available_decisions = get_available_decisions_for_functions(experiment.optimization_graph, arguments)

    # calculate costs for all combinations of decisions
    if arguments.exhaustive:
        best_configuration = evaluate_all_decision_combinations(
            experiment, available_decisions, arguments, optimizer_dir
        )
    elif arguments.evolutionary != None:
        # perform evolutionary search
        best_configuration = perform_evolutionary_search(
            experiment,
            available_decisions,
            arguments,
            optimizer_dir,
        )
    else:
        raise ValueError("No optimization method specified!")

    if best_configuration is not None:
        best_configuration = optimize_updates(experiment, best_configuration, arguments)
        # append the configuration to the list of patterns
        experiment.detection_result.patterns.optimizer_output.append(best_configuration)

    # save full experiment to disk
    export_to_json(experiment, optimizer_dir)
    # save updated patterns.json to disk
    export_patterns_to_json(experiment, os.path.join(optimizer_dir, "patterns.json"))
    # save updated detection_result to disk
    export_detection_result_to_json(experiment, os.path.join(optimizer_dir, "detection_result_dump.json"))



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import os
import re
import warnings
from collections import defaultdict
from dataclasses import dataclass
from os.path import abspath, dirname
from typing import Any, List, Tuple

from lxml import objectify  # type:ignore

# Map to record which line belongs to read set of nodes. LID -> NodeIds
readlineToCUIdMap = defaultdict(set)  # type: ignore
# Map to record which line belongs to write set of nodes. LID -> NodeIds
writelineToCUIdMap = defaultdict(set)  # type: ignore
# Map to record which line belongs to set of nodes. LID -> NodeIds
lineToCUIdMap = defaultdict(set)  # type: ignore


@dataclass
class DependenceItem(object):
    sink: Any
    source: Any
    type: Any
    var_name: Any
    memory_region: Any
    # TODO improve typing


# TODO move this class to a better place, we need it not only for parsing
@dataclass
class LoopData(object):
    line_id: str  # file_id:line_nr
    total_iteration_count: int
    entry_count: int
    average_iteration_count: int
    maximum_iteration_count: int


def __parse_xml_input(xml_fd):
    xml_content = ""
    for line in xml_fd.readlines():
        if not (line.rstrip().endswith("</Nodes>") or line.rstrip().endswith("<Nodes>")):
            xml_content = xml_content + line

    xml_content = "<Nodes>{0}</Nodes>".format(xml_content)

    parsed_cu = objectify.fromstring(xml_content)
    cu_dict = dict()
    for node in parsed_cu.Node:
        node.childrenNodes = str(node.childrenNodes).split(",") if node.childrenNodes else []
        if node.get("type") == "0":
            for instruction_id in str(node.writePhaseLines).split(","):
                writelineToCUIdMap[instruction_id].add(node.get("id"))
            for instruction_id in str(node.readPhaseLines).split(","):
                readlineToCUIdMap[instruction_id].add(node.get("id"))

        if node.get("id") in cu_dict:
            # entry exists already! merge the two entries
            pass
        else:
            cu_dict[node.get("id")] = node

    return cu_dict


def __map_dummy_nodes(cu_dict):
    dummy_node_args_to_id_map = defaultdict(list)
    func_node_args_to_id_map = dict()
    dummy_to_func_ids_map = dict()
    for node_id, node in cu_dict.items():
        if node.get("type") == "3" or node.get("type") == "1":
            key = node.get("name")
            if "arg" in dir(node.funcArguments):
                for i in node.funcArguments.arg:
                    key = key + i.get("type")
            if node.get("type") == "3":
                dummy_node_args_to_id_map[key].append(node_id)
            else:
                func_node_args_to_id_map[key] = node_id

    # iterate over all real functions
    for func in func_node_args_to_id_map:
        if func in dummy_node_args_to_id_map:
            for dummyID in dummy_node_args_to_id_map[func]:
                dummy_to_func_ids_map[dummyID] = func_node_args_to_id_map[func]
                cu_dict.pop(dummyID)

    # now go through all the nodes and update the mapped dummies to real funcs
    for node_id, node in cu_dict.items():
        # check dummy in all the children nodes
        if "childrenNodes" in dir(node):
            for child_idx, child in enumerate(node.childrenNodes):
                if child in dummy_to_func_ids_map:
                    cu_dict[node_id].childrenNodes[child_idx] = dummy_to_func_ids_map[child]

            # Also do the same in callLineToFunctionMap
            if "callsNode" in dir(node):
                for idx, i in enumerate(node.callsNode.nodeCalled):
                    if i in dummy_to_func_ids_map:
                        cu_dict[node_id].callsNode.nodeCalled[idx] = dummy_to_func_ids_map[i]
    return cu_dict


def __parse_dep_file(dep_fd, output_path: str) -> Tuple[List[DependenceItem], List[LoopData]]:
    dependencies_list: List[DependenceItem] = []
    loop_data_list: List[LoopData] = []
    # read static dependencies
    static_dependency_lines = []
    if not os.path.exists(os.path.join(output_path, "static_dependencies.txt")):
        warnings.warn(
            "Static dependencies could not be found under: " + os.path.join(output_path, "static_dependencies.txt")
        )
        # todo
        warnings.warn(
            "TODO: Add command line parameter to pass a location for the static dependency file, "
            "or combine static and dynamic dependencies from the beginning."
        )
    else:
        with open(os.path.join(output_path, "static_dependencies.txt"), "r") as static_dep_fd:
            static_dependency_lines = static_dep_fd.readlines()

    for line in dep_fd.readlines() + static_dependency_lines:
        dep_fields = line.split()
        if dep_fields[1] == "BGN" and dep_fields[2] == "loop":
            line_id = dep_fields[0]
            total_iteration_count = int(dep_fields[3])
            entry_count = int(dep_fields[4])
            average_iteration_count = int(dep_fields[5])
            maximum_iteration_count = int(dep_fields[6])
            loop_data_list.append(
                LoopData(
                    line_id,
                    total_iteration_count,
                    entry_count,
                    average_iteration_count,
                    maximum_iteration_count,
                )
            )
        if len(dep_fields) < 4 or dep_fields[1] != "NOM":
            continue
        sink = dep_fields[0]
        # pairwise iteration over dependencies source
        for dep_pair in list(zip(dep_fields[2:], dep_fields[3:]))[::2]:
            type = dep_pair[0]
            source_fields = dep_pair[1].split("|")
            var_str = "" if len(source_fields) == 1 else source_fields[1]
            var_name = ""
            aa_var_name = ""
            if len(var_str) > 0:
                if "(" in var_str:
                    split_var_str = var_str.split("(")
                    var_name = split_var_str[0]
                    aa_var_name = split_var_str[1][
                        :-1
                    ]  # name of the allocated variable which is accessed, i.e. variable name after anti aliasing
                else:
                    # compatibility with results created without alias analysis
                    var_name = var_str
            dependencies_list.append(DependenceItem(sink, source_fields[0], type, var_name, aa_var_name))

    return dependencies_list, loop_data_list


def parse_inputs(cu_file, dependencies, reduction_file, file_mapping):
    with open(cu_file) as f:
        cu_dict = __parse_xml_input(f)
    cu_dict = __map_dummy_nodes(cu_dict)

    with open(dependencies) as f:
        dependencies, loop_info = __parse_dep_file(f, dirname(abspath(cu_file)))

    loop_data = {loop.line_id: loop for loop in loop_info}

    fmap_file = open(file_mapping)
    fmap_lines = fmap_file.read().splitlines()
    fmap_file.close()

    if os.path.exists(reduction_file):
        reduction_vars = []

        # parse reduction variables
        with open(reduction_file) as f:
            content = f.readlines()
        for line in content:
            if is_reduction(line, fmap_lines, file_mapping):
                line = line.replace("\n", "")
                s = line.split(" ")
                var = {
                    "loop_line": f"{s[3]}:{s[8]}",
                    "name": s[17],
                    "reduction_line": f"{s[3]}:{s[13]}",
                    "operation": s[21],
                }
                reduction_vars.append(var)
    else:
        reduction_vars = None

    return cu_dict, dependencies, loop_data, reduction_vars


def is_reduction(reduction_line, fmap_lines, file_mapping):
    rex = re.compile("FileID : ([0-9]*) Loop Line Number : [0-9]* Reduction Line Number : ([0-9]*) ")
    if not rex:
        return False
    res = rex.search(reduction_line)
    file_id = int(res.group(1))
    file_line = int(res.group(2))

    filepath = get_filepath(file_id, fmap_lines, file_mapping)
    if not os.path.exists(filepath):
        return False
    src_file = open(filepath)
    src_lines = src_file.read().splitlines()
    src_file.close()

    return possible_reduction(file_line, src_lines)


def possible_reduction(line, src_lines):
    assert line > 0 and line <= len(src_lines), "invalid src line"
    src_line = src_lines[line - 1]
    while not ";" in src_line:
        line = line + 1
        if line > len(src_lines):
            return False
        src_line = src_line + " " + src_lines[line - 1]

    pos = src_line.find("=")
    if pos == -1:
        return True

    bracket_a = src_line[0:pos].find("[")
    if bracket_a == -1:
        return True
    bracket_b = src_line[0:pos].rfind("]")
    assert bracket_b != -1

    rex_search_res = re.search("([A-Za-z0-9_]+)\[", src_line[0 : (bracket_a + 1)])
    if not rex_search_res:
        return True

    array_name = rex_search_res[1]
    array_index = src_line[(bracket_a + 1) : bracket_b]

    array_indices = find_array_indices(array_name, src_line[pos : len(src_line)])
    for index in array_indices:
        if index == array_index:
            return False

    return True


def get_filepath(file_id, fmap_lines, file_mapping):
    assert file_id > 0 and file_id <= len(fmap_lines), "invalid file id"
    line = fmap_lines[file_id - 1]
    tokens = line.split(sep="\t")
    if tokens[1].startswith(".."):
        return os.path.dirname(file_mapping) + "/" + tokens[1]
    else:
        return tokens[1]


def get_enclosed_str(data):
    num_open_brackets = 1
    for i in range(0, len(data)):
        if data[i] == "[":
            num_open_brackets = num_open_brackets + 1
        elif data[i] == "]":
            num_open_brackets = num_open_brackets - 1
            if num_open_brackets == 0:
                return data[0:i]


def find_array_indices(array_name, src_line):
    indices = []
    uses = list(re.finditer(array_name, src_line))
    for use in uses:
        if src_line[use.end()] == "[":
            indices.append(get_enclosed_str(src_line[(use.end() + 1) : len(src_line)]))
        else:
            indices.append("")

    return indices



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import tkinter as tk


class TKVarStorage(object):
    # code preview settings
    toggle_var_code_preview_show_metadata_regions: tk.IntVar
    toggle_var_code_preview_show_metadata_live_device_variables: tk.IntVar
    toggle_var_code_preview_show_line_numbers: tk.IntVar
    toggle_var_code_preview_disable_compile_check: tk.IntVar

    def __init__(self, wizard):
        self.wizard = wizard
        self.toggle_var_code_preview_show_metadata_regions = tk.IntVar(
            value=self.wizard.settings.code_preview_show_metadata_regions
        )
        self.toggle_var_code_preview_show_line_numbers = tk.IntVar(
            value=self.wizard.settings.code_preview_show_line_numbers
        )
        self.toggle_var_code_preview_show_metadata_live_device_variables = tk.IntVar(
            value=self.wizard.settings.code_preview_show_metadata_live_device_variables
        )
        self.toggle_var_code_preview_disable_compile_check = tk.IntVar(
            value=self.wizard.settings.code_preview_disable_compile_check
        )

    def toggle_code_preview_setting_action(self):
        """overwrites the respective value in wizard.settings and triggers saving the new settings."""
        self.wizard.settings.code_preview_show_metadata_regions = (
            self.toggle_var_code_preview_show_metadata_regions.get()
        )
        self.wizard.settings.code_preview_show_metadata_live_device_variables = (
            self.toggle_var_code_preview_show_metadata_live_device_variables.get()
        )
        self.wizard.settings.code_preview_show_line_numbers = self.toggle_var_code_preview_show_line_numbers.get()
        self.wizard.settings.code_preview_disable_compile_check = (
            self.toggle_var_code_preview_disable_compile_check.get()
        )
        self.wizard.settings.save_to_file(self.wizard.config_dir)
        self.wizard.console.print("Saved settings.")



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import os
import pathlib
import signal
import tkinter as tk
import warnings
from enum import IntEnum
from os.path import dirname
from tkinter import messagebox, filedialog
from tkinter import ttk
from typing import Optional

from DEPRECATED_discopop_wizard.classes.Arguments import Arguments
from DEPRECATED_discopop_wizard.classes.Console import Console
from DEPRECATED_discopop_wizard.classes.ProfilingContainer import ProfilingContainer
from DEPRECATED_discopop_wizard.classes.Settings import Settings, load_from_config_file
from DEPRECATED_discopop_wizard.classes.TKVarStorage import TKVarStorage
from DEPRECATED_discopop_wizard.screens.main import MainScreen

# todo add command line option to list available run configurations
# todo add command line option to execute run configuration (by name)
from DEPRECATED_discopop_wizard.screens.settings import show_settings_screen
from DEPRECATED_discopop_wizard.utils import get_platform, Platform


def main(arguments: Arguments):
    print("starting DiscoPoP Wizard...\n")
    source_dir = dirname(os.path.abspath(__file__))  # source_dir: discopop/discopop_wizard
    config_dir = os.path.join(source_dir, ".config")

    # check if config exists, if not, initialize config folder and files
    if not os.path.exists(config_dir):
        os.mkdir(config_dir)
    if not os.path.exists(os.path.join(config_dir, "execution_configurations")):
        os.mkdir(os.path.join(config_dir, "execution_configurations"))

    # check if SETTINGS file exists. if not, create it.
    if not os.path.exists(os.path.join(config_dir, "SETTINGS.txt")):
        with open(os.path.join(config_dir, "SETTINGS.txt"), "w+"):
            pass

    wizard = DiscoPoPConfigurationWizard(config_dir, arguments)
    print()


class ConsoleStyles(IntEnum):
    NORMAL = 1
    WARNING = 2
    ERROR = 3


class DiscoPoPConfigurationWizard(object):
    arguments: Arguments
    settings: Settings
    window: tk.Tk
    window_frame: tk.Frame
    config_dir: str
    menubar: tk.Menu
    profiling_container: Optional[ProfilingContainer] = None
    tk_var_storage: TKVarStorage

    ## font styles
    style_font_bold: str = "Helvetica 12 bold"
    style_font_bold_small: str = "Helvetica 10 bold"

    def __init__(self, config_dir: str, arguments: Arguments, headless_mode: bool = False):
        self.arguments = arguments
        self.config_dir = config_dir
        # check if settings exist
        if os.stat(os.path.join(config_dir, "SETTINGS.txt")).st_size == 0:
            # no settings exist
            prompt_result = messagebox.askyesno(
                "DiscoPoP Wizard",
                "Do you want to make use of a docker container for the profiling?",
            )
            if not prompt_result:
                # ask user for path to discopop_build and go/bin directory
                discopop_build_dir = filedialog.askdirectory(title="Select DiscoPoP build folder")
                go_bin_dir = filedialog.askdirectory(title="Select go/bin folder (Go installation)")
            else:
                discopop_build_dir = ""
                go_bin_dir = ""
            self.settings = Settings(
                use_docker_container=prompt_result,
                discopop_build_dir=discopop_build_dir,
                go_bin_dir=go_bin_dir,
            )
        else:
            # load settings
            self.settings = load_from_config_file(config_dir)
        if not headless_mode:
            self.initialize_screen(config_dir)

    def initialize_screen(self, config_dir: str):
        self.window = tk.Tk()
        self.window.title("DiscoPoP Wizard")

        # enable closing by pressing CTRL+C in the command line or the interface
        def handler(event):
            self.window.destroy()
            print("caught ^C")

        def check():
            self.window.after(500, check)

        signal.signal(signal.SIGINT, lambda x, y: print("terminal ^C") or handler(None))
        self.window.after(500, check)
        # self.window.bind_all('<Control-c>', handler) # uncomment to close with CTRL+C from interface

        # load window icon
        try:
            photo = tk.PhotoImage(
                file=os.path.join(
                    str(pathlib.Path(__file__).parent.resolve()),
                    "assets",
                    "icons",
                    "discoPoP_128x128.png",
                )
            )
            self.window.iconphoto(False, photo)
        except tk.TclError:
            warnings.warn("Loading the window icon was not successful.")

        # set window to full screen
        if get_platform() in (Platform.OSX, Platform.WINDOWS):
            self.window.state("zoomed")
        else:
            self.window.attributes("-zoomed", True)
        self.window.columnconfigure(1, weight=1)
        self.window.rowconfigure(1, weight=1)
        paned_window = ttk.PanedWindow(self.window, orient=tk.VERTICAL)
        paned_window.pack(fill=tk.BOTH, expand=True)

        # create content frame
        self.window_frame = tk.Frame(paned_window)
        paned_window.add(self.window_frame, weight=5)
        self.window_frame.columnconfigure(1, weight=1)
        self.window_frame.rowconfigure(1, weight=1)

        # create menu bar
        self.menubar = tk.Menu(self.window)
        self.window.config(menu=self.menubar)

        # create console frame
        self.console_frame = tk.Frame(paned_window)
        paned_window.add(self.console_frame, weight=0)
        self.console_frame.columnconfigure(1, weight=1)
        self.console_frame.rowconfigure(1, weight=1)
        self.console = Console(self.console_frame)

        # create TKVarStorage
        self.tk_var_storage = TKVarStorage(self)

        MainScreen(self, self.window_frame)

        # show settings screen if first start
        if not self.settings.initialized:
            show_settings_screen(self)
        else:
            # save settings
            self.settings.save_to_file(self.config_dir)

        self.window.mainloop()

        # close DiscoPoP profiling profiling_container before exiting the application
        if self.profiling_container is not None:
            self.profiling_container.stop()

    def close_frame_contents(self):
        # close current frame contents
        for c in self.window_frame.winfo_children():
            c.destroy()
        # create empty menu bar
        self.menubar.destroy()
        self.menubar = tk.Menu(self.window)
        self.window.config(menu=self.menubar)

    def show_main_screen(self):
        self.close_frame_contents()
        MainScreen(self, self.window_frame)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import cProfile
import json
import os
import sys
import time
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional

import pstats2  # type:ignore
from pluginbase import PluginBase  # type:ignore

from discopop_library.LineMapping.initialize import initialize_line_mapping
from discopop_library.PathManagement.PathManagement import get_path, load_file_mapping
from discopop_library.discopop_optimizer.Microbench.ExtrapInterpolatedMicrobench import (
    ExtrapInterpolatedMicrobench,
)
from discopop_library.result_classes.DetectionResult import DetectionResult
from .PEGraphX import PEGraphX
from .json_serializer import PatternBaseSerializer
from .parser import parse_inputs
from .pattern_detection import PatternDetectorX


@dataclass
class ExplorerArguments(object):
    """Container Class for the arguments passed to the discopop_explorer"""

    # input files and configuration
    discopop_build_path: str
    project_path: str
    cu_xml_file: str
    dep_file: str
    loop_counter_file: str
    reduction_file: str
    file_mapping_file: str
    plugins: List[str]
    # output and formatting
    enable_json_file: Optional[str]
    enable_profiling_dump_file: Optional[str]  # None means no dump, otherwise the path
    enable_pet_dump_file: Optional[str]  # None means no dump, otherwise the path
    enable_detection_result_dump_file: Optional[str]  # None means no dump, otherwise the path
    enable_patterns: str
    # experimental features:
    enable_task_pattern: bool
    detect_scheduling_clauses: bool
    generate_data_cu_inst: Optional[str]  # none: generate Data_CUInst.txt & exit
    cu_inst_result_file: Optional[str]
    llvm_cxxfilt_path: Optional[str]
    microbench_file: Optional[str]

    def __post_init__(self):
        self.__validate()

    def __validate(self):
        """Validate the arguments passed to the discopop_explorer, e.g check if given files exist"""
        validation_failure = False

        # check for missing files
        missing_files = []
        for file in [
            self.cu_xml_file,
            self.dep_file,
            self.loop_counter_file,
            self.reduction_file,
            self.file_mapping_file,
        ]:  # TODO more files?
            if not os.path.isfile(file):
                missing_files.append(file)
        if missing_files:
            validation_failure = True
            print("The following files are missing:")
            for file in missing_files:
                print(file)

        # validate discopop build path
        # TODO validate discopop build path

        if validation_failure:
            print("Exiting...")
            sys.exit()


def __run(
    project_path: str,
    cu_xml: str,
    dep_file: str,
    loop_counter_file: str,  # TODO we should be able to read all info from the _dep.txt file (?)
    reduction_file: str,
    plugins: List[str],
    file_mapping: Optional[str] = None,
    cu_inst_result_file: Optional[str] = None,
    llvm_cxxfilt_path: Optional[str] = None,
    discopop_build_path: Optional[str] = None,
    enable_patterns: str = "*",
    enable_task_pattern: bool = False,
    enable_detection_of_scheduling_clauses: bool = False,
) -> DetectionResult:
    pet = PEGraphX.from_parsed_input(*parse_inputs(cu_xml, dep_file, reduction_file, file_mapping))
    print("PET CREATION FINISHED.")
    # pet.show()
    # TODO add visualization

    plugin_base = PluginBase(package="plugins")

    plugin_source = plugin_base.make_plugin_source(searchpath=[Path(__file__).parent / "plugins"])

    for plugin_name in plugins:
        p = plugin_source.load_plugin(plugin_name)
        print("executing plugin before: " + plugin_name)
        pet = p.run_before(pet)

    pattern_detector = PatternDetectorX(pet)

    res: DetectionResult = pattern_detector.detect_patterns(
        project_path,
        cu_xml,
        dep_file,
        loop_counter_file,
        reduction_file,
        file_mapping,
        cu_inst_result_file,
        llvm_cxxfilt_path,
        discopop_build_path,
        enable_patterns,
        enable_task_pattern,
        enable_detection_of_scheduling_clauses,
    )

    for plugin_name in plugins:
        p = plugin_source.load_plugin(plugin_name)
        # print("executing plugin after: " + plugin_name)
        pet = p.run_after(pet)

    return res


def run(arguments: ExplorerArguments):
    """Run the discopop_explorer with the given arguments"""
    # create explorer directory if not already present
    if not os.path.exists(os.path.join(arguments.project_path, "explorer")):
        os.mkdir(os.path.join(arguments.project_path, "explorer"))
    # create file to store next free pattern ids if not already present
    if not os.path.exists("next_free_pattern_id.txt"):
        with open("next_free_pattern_id.txt", "w") as f:
            f.write(str(0))
    # reset file lock in case of prior crashes
    if os.path.exists("next_free_pattern_id.txt.lock"):
        os.remove("next_free_pattern_id.txt.lock")

    if arguments.enable_profiling_dump_file is not None:
        profile = cProfile.Profile()
        profile.enable()

    if arguments.generate_data_cu_inst is not None:
        # start generation of Data_CUInst and stop execution afterwards
        from .generate_Data_CUInst import wrapper as generate_data_cuinst_wrapper

        generate_data_cuinst_wrapper(
            arguments.cu_xml_file,
            arguments.dep_file,
            arguments.loop_counter_file,
            arguments.reduction_file,
            arguments.generate_data_cu_inst,
        )
        sys.exit(0)

    start = time.time()

    res = __run(
        arguments.project_path,
        arguments.cu_xml_file,
        arguments.dep_file,
        arguments.loop_counter_file,
        arguments.reduction_file,
        arguments.plugins,
        file_mapping=arguments.file_mapping_file,
        cu_inst_result_file=arguments.cu_inst_result_file,
        llvm_cxxfilt_path=arguments.llvm_cxxfilt_path,
        discopop_build_path=arguments.discopop_build_path,
        enable_patterns=arguments.enable_patterns,
        enable_task_pattern=arguments.enable_task_pattern,
        enable_detection_of_scheduling_clauses=arguments.detect_scheduling_clauses,
    )

    end = time.time()

    if arguments.enable_pet_dump_file is not None:
        with open(arguments.enable_pet_dump_file, "w+") as f:
            f.write(res.pet.dump_to_pickled_json())
            f.flush()
            f.close()

    if arguments.enable_detection_result_dump_file is not None:
        with open(arguments.enable_detection_result_dump_file, "w+") as f:
            f.write(res.dump_to_pickled_json())
            f.flush()
            f.close()

    if arguments.enable_json_file is None:
        print(str(res))
    else:
        # todo re-enable?
        # print(str(res))
        # since PETGraphX is not JSON Serializable, delete the field prior to executing the serialization
        del res.pet
        with open(arguments.enable_json_file, "w+") as f:
            json.dump(res, f, indent=2, cls=PatternBaseSerializer)

    if arguments.enable_profiling_dump_file is not None:
        profile.disable()
        if os.path.exists(arguments.enable_profiling_dump_file):
            os.remove(arguments.enable_profiling_dump_file)
        with open(arguments.enable_profiling_dump_file, "w+") as f:
            stats = pstats2.Stats(profile, stream=f).sort_stats("time").reverse_order()
            stats.print_stats()

    # initialize the line_mapping.json
    initialize_line_mapping(load_file_mapping(arguments.file_mapping_file), arguments.project_path)

    print("Time taken for pattern detection: {0}".format(end - start))

    # demonstration of Microbenchmark possibilities
    if arguments.microbench_file is not None:
        microbench_file = get_path(
            arguments.project_path, arguments.microbench_file
        )  # NOTE: the json file is not usually located in the project, this is just for demonstration purposes
        if not os.path.isfile(microbench_file):
            raise FileNotFoundError(f"Microbenchmark file not found: {microbench_file}")
        extrapBench = ExtrapInterpolatedMicrobench(microbench_file)
        sympyExpr = extrapBench.getFunctionSympy()
        print(sympyExpr)
        print(sympyExpr.free_symbols)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import json
import os.path
from pathlib import Path
import shutil
from typing import Dict, List, Optional, Tuple
from discopop_library.CodeGenerator.CodeGenerator import from_json_strings, from_json_strings_with_mapping
from discopop_library.PatchGenerator.PatchGeneratorArguments import PatchGeneratorArguments
from discopop_library.PatchGenerator.diffs import get_diffs_from_modified_code
from discopop_library.discopop_optimizer.classes.context.Update import Update, construct_update_from_dict
from discopop_library.discopop_optimizer.classes.system.devices.DeviceTypeEnum import DeviceTypeEnum
from discopop_library.discopop_optimizer.classes.types.Aliases import DeviceID
from discopop_library.result_classes.DetectionResult import DetectionResult
import jsonpickle  # type: ignore

from discopop_library.result_classes.OptimizerOutputPattern import OptimizerOutputPattern  # type: ignore


def from_optimizer_output(
    file_mapping: Dict[int, Path],
    patterns_by_type: Dict[str, List[str]],
    optimizer_output_json_str: str,
    arguments: PatchGeneratorArguments,
    patch_generator_dir: str,
):
    suggestion_strings_with_mapping: Dict[str, List[Tuple[str, DeviceID, Optional[DeviceTypeEnum]]]] = dict()
    if arguments.verbose:
        print("Loading optimizer output: ")

    optimizer_output = json.loads(optimizer_output_json_str)

    # load detectionresult and pet
    if arguments.verbose:
        print("Loading detection result and PET...", end="")
    detection_result_dump_str = ""
    with open(os.path.join("explorer", "detection_result_dump.json")) as f:
        detection_result_dump_str = f.read()
    detection_result: DetectionResult = jsonpickle.decode(detection_result_dump_str)
    if arguments.verbose:
        print("Done")

    # build suggestion_strings_with_mapping
    for pattern_values in optimizer_output["applied_patterns"]:
        pattern_id = pattern_values["pattern_id"]
        device_id = pattern_values["device_id"]
        device_type = pattern_values["device_type"]
        for pattern_type in patterns_by_type:
            if pattern_type == "version":
                continue
            for pattern_string in patterns_by_type[pattern_type]:
                loaded_pattern = json.loads(pattern_string)
                if loaded_pattern["pattern_id"] == pattern_id:
                    if pattern_type not in suggestion_strings_with_mapping:
                        suggestion_strings_with_mapping[pattern_type] = []
                    suggestion_strings_with_mapping[pattern_type].append((pattern_string, device_id, device_type))

    # collect data movement information
    if "device_update" not in suggestion_strings_with_mapping:
        suggestion_strings_with_mapping["device_update"] = []
    for data_movement in optimizer_output["data_movement"]:
        update_obj = construct_update_from_dict(data_movement)
        suggestion_strings_with_mapping["device_update"].append(
            (update_obj.get_pattern_string(detection_result.pet), None, None)
        )

    # generate the modified code
    file_id_to_modified_code: Dict[int, str] = from_json_strings_with_mapping(
        file_mapping,
        suggestion_strings_with_mapping,
        CC=arguments.CC,
        CXX=arguments.CXX,
        skip_compilation_check=True,
        host_device_id=optimizer_output["host_device_id"],
    )
    print("MODIFIED CODE: ")
    print(file_id_to_modified_code)

    # create patches from the modified codes
    file_id_to_patches: Dict[int, str] = get_diffs_from_modified_code(file_mapping, file_id_to_modified_code, arguments)
    if arguments.verbose:
        print("Patches: ", file_id_to_patches)
    # save patches
    suggestion_id = optimizer_output["pattern_id"]

    suggestion_folder_path = os.path.join(patch_generator_dir, str(suggestion_id))
    if arguments.verbose:
        print("Saving patches for suggestion: ", suggestion_id)
    if os.path.exists(suggestion_folder_path):
        shutil.rmtree(suggestion_folder_path)
    os.mkdir(suggestion_folder_path)
    for file_id in file_id_to_patches:
        patch_path = os.path.join(suggestion_folder_path, str(file_id) + ".patch")
        with open(patch_path, "w") as f:
            f.write(file_id_to_patches[file_id])
